# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''此模块用于归纳基础通用子模块
'''

import re
import sys
import time
import subprocess
import threading
from six import StringIO
from six import PY3

class Task(object):
    
    def __init__(self, command, cwd=None, env=None, print_enable=False):
        '''构造函数
        @param command: 命令行
        @type command: <str>
        @param cwd: 设置执行目录（默认为父进程目录）
        @type cwd: <str>
        @param env: 设置环境变量
        @type env: <dict>
        @param print_enable: 开启print
        @type print_enable: <bool>
        '''
        self.command = command
        self.bufsize = 1
        self.shell = True
        self.cwd = cwd
        self.env = env
        self.print_enable = print_enable
    
    def __execute__(self, out, err):
        '''使用subprocess模块执行
        '''
        process = subprocess.Popen(self.command, bufsize=self.bufsize, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                    shell=self.shell, cwd=self.cwd, env=self.env)
        while True:
            out_line = process.stdout.readline()
            if PY3:
                out_line = out_line.decode()
            if out_line == '' and process.poll() != None    : break
            if out_line != '' and self.print_enable == True : sys.stdout.write(out_line)

            out.writelines([out_line])
            out.flush()
        while True:
            err_line = process.stderr.readline()
            if PY3:
                err_line = err_line.decode()
            if err_line == '' and process.poll() != None    : break
            if err_line != '' and self.print_enable == True : sys.stderr.write(err_line)
            err.writelines([err_line])
            err.flush()
        out.seek(0)
        err.seek(0)
        return out, err, process.returncode
    
    def execute1(self):
        '''执行命令
        @return: stdout and stderr<StringIO>
        '''
        out = StringIO()
        self.__execute__(out, out)
        return out
    
    def execute2(self):
        '''执行命令
        @return: (stdout<StringIO>, stderr<StringIO>)
        '''
        out, err = StringIO(), StringIO()
        self.__execute__(out, err)
        return out, err
    
    def execute3(self):
        '''执行命令
        @return: (stdout<StringIO>, stderr<StringIO>, returncode<int>)
        '''
        out, err = StringIO(), StringIO()
        returncode = self.__execute__(out, err)[-1]
        return out, err, returncode
    
    def execute(self):
        '''执行命令
        @return: stdout<str>
        '''
        out = self.execute1()
        return re.sub("\n+$", "", out.read())


class ThreadTask(threading.Thread):
    '''该线程类用于执行Shell
    '''
    
    def __init__(self, command, started_callback=None, stdout_line_callback=None, stderr_line_callback=None, return_code_callback=None, cwd=None, env=None, print_enable=False):
        '''构造函数
        @param command: 命令行
        @type command: <str>
        @param started_callback: started回调
        @type started_callback: <function> 
        @param stdout_line_callback: stdout行回调
        @type stdout_line_callback: <function>
        @param stderr_line_callback: stderr行回调
        @type stderr_line_callback: <function>
        @param return_code_callback: returncode回调
        @type return_code_callback: <function>
        @param cwd: 设置执行目录（默认为父进程目录）
        @type cwd: <str>
        @param env: 设置环境变量
        @type env: <dict>
        @param print_enable: 开启print
        @type print_enable: <bool>
        '''
        threading.Thread.__init__(self)
        self.command = command
        self.bufsize = 1
        self.shell = True
        self.started_callback = started_callback
        self.stdout_line_callback = stdout_line_callback
        self.stderr_line_callback = stderr_line_callback
        self.return_code_callback = return_code_callback
        self.cwd = cwd
        self.env = env
        self.print_enable = print_enable
        self.process = None
        self.keep_running = True
        self.setDaemon(True)
        self.start()
    
    def __call_started_callback__(self, pid):
        '''回调started
        '''
        if self.started_callback and callable(self.started_callback):
            self.started_callback(pid)
    
    def __call_stdout_Line_callback__(self, line):
        '''回调stdout
        '''
        if self.print_enable : sys.stdout.write(line)
        if PY3:
            line = line.decode()
        if self.stdout_line_callback and callable(self.stdout_line_callback):
            self.stdout_line_callback(line)
    
    def __call_stderr_line_callback__(self, line):
        '''回调stderr
        '''
        if self.print_enable : sys.stderr.write(line)
        if PY3:
            line = line.decode()
        if self.stderr_line_callback and callable(self.stderr_line_callback):
            self.stderr_line_callback(line)
    
    def __call_return_code_callback__(self, pid, returncode):
        '''回调returncode
        '''
        if self.return_code_callback and callable(self.return_code_callback):
            self.return_code_callback(pid, returncode)
    
    @property
    def pid(self):
        '''获取进程PID
        '''
        if self.process: return self.process.pid
    
    def poll(self):
        '''进程是否执行完毕
        '''
        if self.process: return self.process.poll()
    
    def is_alive(self):
        '''进程是否还存在
        '''
        if hasattr(self.process, 'returncode'):
            return self.process.returncode == None
        return False
    
    def terminate(self):
        '''终止进程
        '''
        if self.is_alive(): self.process.terminate()
    
    def kill(self):
        '''杀进程
        '''
        if self.is_alive(): self.process.kill()
    
    def run(self):
        '''线程开始执行（该方法自动执行无需调用start）
        '''
        self.process = subprocess.Popen(self.command, bufsize=self.bufsize, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=self.shell, cwd=self.cwd, env=self.env)
        self.__call_started_callback__(self.process.pid)
        while self.keep_running:
            out_line = self.process.stdout.readline()
            if out_line == '' and self.process.poll() != None : break
            if out_line != '' : self.__call_stdout_Line_callback__(out_line)
        while self.keep_running:
            err_line = self.process.stderr.readline()
            if err_line == '' and self.process.poll() != None : break
            if err_line != '' : self.__call_stderr_line_callback__(err_line)
        self.__call_return_code_callback__(self.process.pid, self.process.returncode)
    
    def stop(self):
        '''终止执行
        '''
        self.keep_running = False
        self.terminate()
    
    def wait(self, timeout=0):
        '''等待至进程结束(外部不要使用join等待线程执行完毕)
        '''
        start_time = time.time()
        while self.poll() == None and self.keep_running:
            if timeout > 0 and time.time() - start_time > timeout : break
            time.sleep(0.3)
        time.sleep(1)



# -*- coding:utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''iOS WebInspector API Driver
'''

from __future__ import absolute_import, print_function

from qt4i.driver.rpc import rpc_method
from qt4i.driver.rpc import RPCEndpoint
from qt4i.driver.tools.dt import DT
from qt4i.driver.tools.logger import get_logger
from qt4i.driver.web.wkrdp import RealDeviceProtocol
from qt4i.driver.web.wkrdp import SimulatorProtocol


class WebInspector(RPCEndpoint):
    '''
    iOS Web Inpector API, 提供iOS的webview的javascript注入接口
    '''
    rpc_name_prefix = "web."
    WKRDP = {}
    WKRDP_SEPARATOR = ':'
    
    def __init__(self, rpc_server, device_id):
        self.udid = device_id
        self.logger = get_logger("driverserver_%s" % device_id)
        RPCEndpoint.__init__(self)

    @rpc_method
    def get_page_id(self, bundle_id, title, url):
        ub = self.WKRDP_SEPARATOR.join([self.udid,bundle_id])
        self.logger.info('current app sessions:%s' % WebInspector.WKRDP)
        if ub not in WebInspector.WKRDP:
            if not DT().is_simulator(self.udid):
                WebInspector.WKRDP[ub] = RealDeviceProtocol(bundle_id, self.udid)
            else:
                WebInspector.WKRDP[ub] = SimulatorProtocol(bundle_id, self.udid)
            self.logger.info('[%s] wkrdp session starting' % ub)
            WebInspector.WKRDP[ub].start()
        return WebInspector.WKRDP[ub].request_current_page_id(title, url)
    
    @rpc_method
    def eval_script(self, bundle_id, frame_id, page_id, script):
        ub = self.WKRDP_SEPARATOR.join([self.udid,bundle_id])
        data = {"method":"Runtime.evaluate","params":{"expression":script,"objectGroup":"console","includeCommandLineAPI":True,"doNotPauseOnExceptionsAndMuteConsole":False,"returnByValue":False,"generatePreview":True,"saveResult":False}}
        WebInspector.WKRDP[ub].send_webkit_socket_data(data, page_id)
        result = self.WKRDP[ub].recv_webkit_socket_data()
        return result['result']['result']
    
    @rpc_method
    def release(self):
        for ub in list(self.WKRDP.keys()):
            if ub.split(self.WKRDP_SEPARATOR)[0] == self.udid:
                self.WKRDP[ub].stop()
                del self.WKRDP[ub]
    
    @rpc_method
    def release_app_session(self, bundle_id):
        ub = WebInspector.WKRDP_SEPARATOR.join([self.udid,bundle_id])
        self.logger.info('release app session:%s' % ub)
        if ub in WebInspector.WKRDP:
            WebInspector.WKRDP[ub].stop()
            del WebInspector.WKRDP[ub]  



# -*- coding:utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''iOS WebKit Remote debug Protocol
'''

from __future__ import absolute_import, print_function

import json
import re
import struct
import socket
import subprocess
import uuid
from past.builtins import xrange
from six import BytesIO
from six import PY3
from six import PY2

from biplist import Data
from biplist import readPlist
from biplist import writePlist

from pymobiledevice.lockdown import LockdownClient
from qt4i.driver.tools.logger import get_logger
from qt4i.driver.tools.sched import PortManager
from qt4i.driver.tools.dt import DT


class EnumSelector(object):
    '''
    iOS Web Inspector消息类型（包含发送消息、接收消息）
    '''
    
    #发送消息类型
    SEND_REPORT_ID = "_rpc_reportIdentifier:"
    SEND_FORWARD_GET_LISTING = "_rpc_forwardGetListing:"
    SEND_FORWARD_SOCKET_SETUP = "_rpc_forwardSocketSetup:"
    SEND_FORWARD_SOCKET_DATA = "_rpc_forwardSocketData:"
    SEND_INDICATE_WEBVIEW = "_rpc_forwardIndicateWebView:"
    
    #接收消息类型
    ON_REPORT_SETUP = "_rpc_reportSetup:"
    ON_REPORT_DRIVER_LIST = "_rpc_reportConnectedDriverList:"
    ON_REPORT_CONNECTED_APP_LIST = "_rpc_reportConnectedApplicationList:"
    ON_APP_CONNECTED = "_rpc_applicationConnected:"
    ON_APP_UPDATED = "_rpc_applicationUpdated:"
    ON_APP_SENT_LISTING = "_rpc_applicationSentListing:"
    ON_APP_SENT_DATA = "_rpc_applicationSentData:"
    ON_APP_DISCONNECTED = "_rpc_applicationDisconnected:"


class WebInspectorError(Exception):
    pass


class WIPageUpdateError(Exception):
    """Web页面发生id更新"""
    pass


class WIPageNotFoundError(Exception):
    """Web页面未找到page id"""
    pass


class WebKitRemoteDebugProtocol(object):
    '''
    iOS WebKit 远程调试协议
    '''

    MAX_PLIST_LEN = 8192 - 512  # 7.5k
    FINAL_MSG = "WIRFinalMessageKey"
    PARTIAL_MSG = "WIRPartialMessageKey"

    def __init__(self, bundle_id, udid=None):
        self.app_bundle_id = bundle_id
        self.udid = udid
        if self.udid:
            self.logger = get_logger("driverserver_%s" % self.udid)
        else:
            self.logger = get_logger()
        self.seq = 0  # webkit调试命令的序号
        self.conn_id = str(uuid.uuid1()).upper()
        self.sender_key = str(uuid.uuid4()).upper()
        self.retry = 5  # 初始化消息的重试次数
        self.msgbuf = []
        self.cached_pages = []
        self.app_id = None
        self.page_id = None
        self.host_app_ids = []
    
    @property
    def is_simulator(self):
        raise NotImplementedError    
        
    def start(self):
        '''启动调试服务
        '''
        raise NotImplementedError
    
    def stop(self):
        '''停止调试服务
        '''
        if hasattr(self, '_sock'):
            self._sock.close()
        
    def send(self, data):
        total = 0
        while total < len(data):
            sent = self._sock.send(data[total:])
            if sent == 0:
                raise WebInspectorError("socket connection (send) error")
            total = total + sent
            
    def recv(self, size=4):
        data = ''
        if PY3:
            data = b''
        while len(data) < size:
            chunk = self._sock.recv(size-len(data))
            if not chunk or len(chunk) == 0:
                raise WebInspectorError("socket connection (recv) error")
            data = data + chunk
        return data     
    
    def _plist_to_bin(self, plist):
        f = BytesIO()
        writePlist(plist, f)
        return f.getvalue()
    
    def _plist_from_bin(self, data):
        f = BytesIO(data)
        return readPlist(f)
    
    def _init_app_id(self):
        request = {'__argument': {'WIRConnectionIdentifierKey': self.conn_id}, 
                   '__selector': EnumSelector.SEND_REPORT_ID}
        self.send_webkit_message(request)
        for _ in xrange(self.retry):
            response = self.recv_webkit_response()
            self.logger.info("wkrdp response:%s" % response)
            if response['__selector'] == EnumSelector.ON_REPORT_SETUP:
                continue
            elif response['__selector'] == EnumSelector.ON_REPORT_CONNECTED_APP_LIST:
                apps_dict = response['__argument']['WIRApplicationDictionaryKey']
                for k in apps_dict:
                    app_info = apps_dict[k]
                    app_bundle_id = app_info['WIRApplicationBundleIdentifierKey']
                    app_id = app_info['WIRApplicationIdentifierKey']
                    
                    if 'WIRHostApplicationIdentifierKey' in app_info: # web页面内嵌在一个app
                        host_app_id = app_info['WIRHostApplicationIdentifierKey']
                        host_app_bundle_id = apps_dict[host_app_id]['WIRApplicationBundleIdentifierKey']
                        if self.app_bundle_id == host_app_bundle_id :
                            self.host_app_ids.append(app_id)
                    if self.app_bundle_id == app_bundle_id:
                        self.app_id = app_id
                        self.app_name = app_info['WIRApplicationNameKey']
                if self.app_id:
                    return
            elif response['__selector'] == EnumSelector.ON_APP_CONNECTED: #单个app信息上报
                app_info = response['__argument']
                if self.app_bundle_id == app_info['WIRApplicationBundleIdentifierKey']:
                    self.app_id = app_info['WIRApplicationIdentifierKey']
                    self.app_name = app_info['WIRApplicationNameKey']
                    return 
        raise Exception('[bundle_id:%s] app id not found' % self.app_bundle_id)

    def _sort_page_ids(self, pages):
        '''排序规则--最新出现的pageID放到list末尾
        '''
        new_cached_pages = [p for p in self.cached_pages if p in pages]
        self.cached_pages = [p for p in pages.keys() if p not in new_cached_pages]
        new_cached_pages.extend(self.cached_pages)
        self.cached_pages = new_cached_pages

    def _get_default_page_id(self, pages):
        '''取最近更新的page作为当前page id
        '''
        new_cached_pages = [p for p in self.cached_pages if p in pages]
        self.cached_pages = [p for p in pages.keys() if p not in new_cached_pages]
        new_cached_pages.extend(self.cached_pages)
        self.cached_pages = new_cached_pages
        ################################ 排除同一个页面空的Webview
        for p in new_cached_pages:
            if 'WIRURLKey' not in pages[p].keys() or \
                (pages[p]['WIRURLKey'].startswith('applewebdata') and pages[p]['WIRTitleKey'] == ''):
                self.cached_pages.remove(p)
                self.cached_pages.insert(0, p) #仅将空page移到队列最前面，保证page列表后续不会进入循环更新
        ################################
        self.logger.info('cached_pages: %s'  %  self.cached_pages)
        page_id = self.cached_pages[-1]
        self.logger.info('page id: %s'  %  self.page_id)
        self.logger.info('new page id: %s'  %  page_id)
        self.title = pages[page_id]['WIRTitleKey']
        self.logger.info('title: %s'  % self.title.encode('utf-8'))
        self.url = pages[page_id]['WIRURLKey']
        self.logger.info('url: %s'  % self.url)
        return page_id
    
    def _get_page_id_by_title(self, pages, title):
        self.logger.info("wkrdp:get_page_id_by_title [%s] of pages" % title)
        self._sort_page_ids(pages)
        for page in pages:
            if 'WIRTitleKey' in pages[page].keys() and pages[page]['WIRTitleKey'] == title:
                self.url = pages[page]['WIRURLKey']
                return page

    def _get_page_id_by_url(self, pages, url):
        self.logger.info("wkrdp:get_page_id_by_url [%s] of pages" %url)
        self._sort_page_ids(pages)
        for page in pages:
            if 'WIRURLKey' in pages[page].keys() and re.search(url, pages[page]['WIRURLKey']):
                self.url = pages[page]['WIRURLKey']
                return page

    def request_current_page_id(self, title=None, url=None):
        if self.app_id not in self.host_app_ids:
            self.host_app_ids.insert(0, self.app_id)
        self.logger.debug('host apps:%s' % self.host_app_ids)
        for app_id in self.host_app_ids:
            for _ in xrange(3):
                request = {'__argument': {'WIRApplicationIdentifierKey': app_id, 
                                          'WIRConnectionIdentifierKey': self.conn_id}, 
                           '__selector': EnumSelector.SEND_FORWARD_GET_LISTING}
                self.send_webkit_message(request)
                try:
                    response = self.recv_webkit_response()
                except socket.timeout:
                    self.logger.error('_rpc_forwardGetListing: timeout')
                    continue
                self.logger.info("page response: %s" % response)
                if response['__selector'] == EnumSelector.ON_REPORT_DRIVER_LIST:
                    try:
                        response = self.recv_webkit_response()
                        self.logger.info("page multi-response: %s" % response)
                    except socket.timeout:
                        continue
                if response['__selector'] == EnumSelector.ON_APP_CONNECTED:  # appID发生变更
                    app_info = response['__argument']
                    if 'WIRHostApplicationIdentifierKey' in app_info and \
                        app_info['WIRHostApplicationIdentifierKey'] == self.app_id and \
                        app_info['WIRIsApplicationProxyKey'] :
                        new_app_id = app_info['WIRApplicationIdentifierKey']
                        if new_app_id not in self.host_app_ids:
                            self.logger.info("add new app id: %s" % new_app_id)
                            self.host_app_ids.append(new_app_id)
                elif response['__selector'] == EnumSelector.ON_APP_SENT_LISTING:
                    res_app_id = response['__argument']['WIRApplicationIdentifierKey']
                    pages = response['__argument']['WIRListingKey']
                    
                    if res_app_id in self.host_app_ids or title or url:
                        if not pages: continue
                        if title:
                            page_id = self._get_page_id_by_title(pages, title)
                        elif url:
                            page_id = self._get_page_id_by_url(pages, url)
                        else:
                            page_id = self._get_default_page_id(pages)
                        if page_id:
                            if res_app_id != self.app_id:
                                self.page_id = None
                            self.app_id = res_app_id
                            return page_id
                elif response['__selector'] == EnumSelector.ON_APP_UPDATED:
                    if response['__argument']['WIRApplicationBundleIdentifierKey'] == self.app_bundle_id:
                        self.app_id = response['__argument']['WIRApplicationIdentifierKey'] 
                elif response['__selector'] == EnumSelector.ON_APP_DISCONNECTED:
                    if self.app_id == response['__argument']['WIRApplicationIdentifierKey']:
                        break
        raise WIPageNotFoundError('[%s]' % self.app_id)

    def _setup_webkit_socket(self, page_id):
        self.page_id = page_id
        request = {'__argument': {'WIRAutomaticallyPause': False, 
                                  'WIRSenderKey': self.sender_key, 
                                  'WIRApplicationIdentifierKey': self.app_id, 
                                  'WIRConnectionIdentifierKey': self.conn_id, 
                                  'WIRPageIdentifierKey': int(page_id)
                                  }, 
                   '__selector': EnumSelector.SEND_FORWARD_SOCKET_SETUP}
        self.send_webkit_message(request)   
        
    def _enable_inspector(self):
        '''打开web inspector的开关(非必须)
        '''  
        data = {"method":"Inspector.enable"}
        self.send_webkit_socket_data(data, self.page_id)
        self.logger.info(self.recv_webkit_socket_data())
        
        data = {"method":"Console.enable"}
        self.send_webkit_socket_data(data, self.page_id)
        self.logger.info(self.recv_webkit_socket_data())  
        
    def init_inspector(self):
        self._init_app_id()
         
    def send_webkit_message(self, message): 
        self.logger.info('='*80)
        self.logger.info("request:%s" % message)
        data = self._plist_to_bin(message)
        data_len = len(data)
        for i in xrange(0, data_len, self.MAX_PLIST_LEN):
            is_partial = data_len - i > self.MAX_PLIST_LEN
            if self.is_simulator:
                tmp_data = data
            else:
                if is_partial:
                    partial_msg = data[i:i + self.MAX_PLIST_LEN]
                    if PY2:
                        partial_msg = Data(partial_msg)
                    else:
                        partial_msg = partial_msg.encode()
                    tmp_data = {self.PARTIAL_MSG: partial_msg}
                else:
                    final_msg = data[i:data_len - i]
                    if PY2:
                        final_msg = Data(final_msg)
                    else:
                        final_msg = final_msg.encode()
                    tmp_data = {self.FINAL_MSG: final_msg}
                tmp_data = self._plist_to_bin(tmp_data)
            tmp_data = struct.pack('!L', len(tmp_data)) + tmp_data
            self.send(tmp_data) 

    def recv_webkit_response(self):
        data = self.recv(4)
        if not data or len(data) != 4:
            raise Exception("WKDP header is invalid")
        l = struct.unpack('!L', data)[0]
        data = self.recv(l)
        data = self._plist_from_bin(data)
        if self.is_simulator:
            return data
        if self.FINAL_MSG in data:
            data = data[self.FINAL_MSG]
            if self.msgbuf:
                data = ''.join(self.msgbuf) + data
                self.msgbuf = []
            data = self._plist_from_bin(data)
            return data 
        else:
            data = data[self.PARTIAL_MSG]
            self.msgbuf.append(data)
            return self.recv_webkit_response()
    
    def send_webkit_socket_data(self, data, page_id):
        if page_id != self.page_id:
            self._setup_webkit_socket(page_id)
        self.seq += 1
        data["id"] = self.seq
        data = json.dumps(data)
        if PY2:
            wrapped_data = Data(data)
        else:
            wrapped_data = data.encode()
        data = {'__argument': {'WIRSenderKey': self.sender_key,
                               'WIRConnectionIdentifierKey': self.conn_id, 
                               'WIRApplicationIdentifierKey': self.app_id, 
                               'WIRSocketDataKey': wrapped_data,
                               'WIRPageIdentifierKey': int(page_id)
                               }, 
                '__selector': EnumSelector.SEND_FORWARD_SOCKET_DATA}
        self.send_webkit_message(data)
    
    def recv_webkit_socket_data(self):
        pages_changed = False
        while True:
            try:
                response = self.recv_webkit_response()
            except socket.timeout:
                if pages_changed:
                    raise WIPageUpdateError('Retry to get new page id.')
                else:
                    raise
            self.logger.info('response: %s' % response)
            selector = response['__selector']
            if selector == EnumSelector.ON_APP_SENT_LISTING:
                self.logger.info("Pages may change!")
                pages_changed = True
                continue
            elif selector != EnumSelector.ON_APP_SENT_DATA or \
                response['__argument']['WIRApplicationIdentifierKey'] != self.app_id:
                continue
            data = response['__argument']['WIRMessageDataKey']
            data = json.loads(data.decode('UTF-8'), encoding='UTF-8')
            if 'id' in data and (data['id'] == self.seq):
                return data


class RealDeviceProtocol(WebKitRemoteDebugProtocol): 
    '''
    iOS真机的远程调试协议实现（依赖usbmuxd通信）
    '''
    
    
    def __init__(self, bundle_id='com.apple.WebKit.WebContent', udid=None):
        super(RealDeviceProtocol, self).__init__(bundle_id, udid)
        self.partial_supported = True
        
    @property
    def is_simulator(self): 
        return not self.partial_supported
        
    def start(self):
        lockdown = LockdownClient(self.udid)
        ios_version = lockdown.allValues['ProductVersion']
        if DT.compare_version(ios_version, '11.0') >= 0: # iOS11真机以上不支持数据分割
            self.partial_supported = False
        self.web_inspector = lockdown.startService("com.apple.webinspector")
        self._sock = self.web_inspector.s
        self._sock.setblocking(False)
        self._sock.settimeout(5)
        self.init_inspector()
        
    def __del__(self):
        self.stop()


class SimulatorProtocol(WebKitRemoteDebugProtocol):
    '''
    iOS模拟器的远程调试协议实现(本地socket通信)
    '''
    
    def __init__(self, bundle_id='com.apple.WebKit.WebContent', udid=None):
        super(SimulatorProtocol, self).__init__(bundle_id, udid)
        if DT.compare_xcode_version('9.3') >= 0:
            self._sock = self._get_webinspector_socket_above_xcode93(udid)
        else:
            self._sock = self._get_webinspector_socket_below_xcode93(udid)
        self._sock.setblocking(False)
        self._sock.settimeout(3)

    def _get_webinspector_socket_above_xcode93(self, udid):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        cmd = ["xcrun", "simctl", "getenv", udid, "RWI_LISTEN_SOCKET"]
        webinspector_unix_sock = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        if PY3:
            webinspector_unix_sock = webinspector_unix_sock.decode()
        webinspector_unix_sock = webinspector_unix_sock.strip('\n')
        self.logger.debug('webinspector socket:%s' % webinspector_unix_sock)
        #'/private/tmp/com.apple.launchd.aVC7CJkaJq/com.apple.webinspectord_sim.socket'
        sock.connect(webinspector_unix_sock)
        return sock

    def _get_webinspector_socket_below_xcode93(self, udid):
        sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
        wi_port = PortManager.get_port('web', udid)
        if wi_port is None:
            wi_port = 27753
        self.logger.info('webport: %s'  %  wi_port)
        sock.connect(('::1', wi_port))
        return sock
        
    @property
    def is_simulator(self): 
        return True
        
    def start(self):
        self.init_inspector()
        
    def __del__(self):
        self.stop()



# -*- coding:utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''QPath
'''


import re, sys


class QPath(str):
    '''QPath基类
    '''
    
    def __init__(self, qpath=None):
        '''构造函数
        
        :param qpath: QPath字符串
        :type qpath: string
        '''
        # -*-
        pattern = '|'.join(['＝', '～', '！'])
        result = re.search(pattern, qpath, re.I)
        if result : 
            sys.stderr.write('警告: QPath中含有中文关系运算符 [%s] in [ %s ]\n' % (result.group(), qpath))
        self._qpath = qpath


    def __str__(self):
        '''获取qpath字符串，实例化控件时调用server.find时传参为qpath字符串
        '''
        return self._qpath


# -*- coding:utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''logging日志模块的封装
'''

import os
import logging
import datetime
import time
import re

from logging.handlers import BaseRotatingHandler
from testbase.conf import settings

TMP_DIR_PATH = settings.get('QT4I_TMP_DIR_PATH', '/tmp')
LOG_MAX_BTYES = 1024*1024*10 #单份日志最大容量(byte)
LOG_BACKUP_COUNT = settings.get('QT4I_LOG_BACKUP_COUNT', 10) #setting获取count时默认值
LOG_DELAY_DAY = 2 #日志保留天数0.0015625
TIME_FORMAT = '%Y-%m-%d %H:%M:%S'
USER_KEY_WORDS = settings.get('QT4I_AGENT_ERROR_KEY_WORDS', [])
DEFAULT_KEY_WORDS = ['requires a development team','invalid code signature','expired','Connection peer refused channel request']

class RotatingFileHandler(BaseRotatingHandler):
    '''文件滚动日志
    '''
    def __init__(self, filename, mode='a', encoding=None, delay=0):
        self._shouldRoller = False
        BaseRotatingHandler.__init__(self, filename, mode, encoding, delay)

    def doRollover(self):
        '''进行日志滚动
        '''
        if self.stream:
            self.stream.close()
            self.stream = None
        if os.path.exists(self.baseFilename):
            os.remove(self.baseFilename)
        if not self.delay:
            self.stream = self._open()
        self._shouldRoller = False

    def shouldRollover(self, record):
        '''决定是否进行日志滚动
        '''
        if self.stream is None:                 # delay was set...
            self.stream = self._open()
        if self._shouldRoller:                   # are we rolling over?
            return 1
        return 0
    
    def rotate(self):
        '''滚动日志
        '''
        self._shouldRoller = True

def get_logger(logger_name='driverserver'):
    '''通过日志名称获取日志对象
    
    :param logger_name: 日志名称
    :type logger_name: str
    :returns: logger_instance - 日志对象
    '''
    logger_instance = logging.getLogger(logger_name)
    
    if len(logger_instance.handlers) == 0:
        logger_instance.setLevel(logging.DEBUG)
        if not os.path.isdir(TMP_DIR_PATH):
            os.mkdir(TMP_DIR_PATH)
        log_path = os.path.join(TMP_DIR_PATH,"%s.log" % logger_instance.name)
        
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.WARNING)

        fmt = "[%(asctime)s][pid:%(process)d][tid:%(thread)d][%(levelname)s] %(message)s"
        formatter = logging.Formatter(fmt)
        formatter2 = logging.Formatter(fmt, TIME_FORMAT)
        console_handler.setFormatter(formatter2)
        
        hdlr = logging.handlers.RotatingFileHandler(log_path, maxBytes=LOG_MAX_BTYES, backupCount=LOG_BACKUP_COUNT)
        hdlr.setLevel(logging.DEBUG)
        hdlr.setFormatter(formatter)
        
        logger_instance.addHandler(console_handler)
        logger_instance.addHandler(hdlr)
        
    return logger_instance

def rotate_logger_by_name(logger_name):
    '''通过日志名称滚动日志，并返回日志对象
    
    :param logger_name: 日志名称
    :type logger_name: str
    :returns: logger_instance - 日志对象
    '''
    logger_instance = get_logger(logger_name)
    for handler in logger_instance.handlers:
        if type(handler) == RotatingFileHandler:
            handler.rotate()
    return logger_instance

def get_logger_path_by_name(logger_name):
    '''通过日志名称获取日志路径
    
    :param logger_name: 日志名称
    :type logger_name: str
    :returns: str or None - 日志路径
    '''
    logger_instance = logging.getLogger(logger_name)
    logger_path = None
    attr_name = 'baseFilename'
    for handler in logger_instance.handlers:
        if hasattr(handler, attr_name):
            logger_path = getattr(handler, attr_name)
            break
    return logger_path

def clean_expired_log(logger_name):
    '''
    根据LOG_DELAY_DAY来清理过期的日志
    
    :param logger_name :日志名称
    :type str
    :return None
    '''
    few_day_ago = (datetime.datetime.now() - datetime.timedelta(days = LOG_DELAY_DAY))
    expire_time = time.mktime(few_day_ago.timetuple())
    files = os.listdir(TMP_DIR_PATH)
    for f in files:
        logger_path = os.path.join(TMP_DIR_PATH, f)
        if os.path.exists(logger_path) and f.find(logger_name) != -1:
            c_time = os.path.getctime(logger_path) #获取文件创建时间
            if c_time < expire_time:
                os.remove(logger_path)

def str_to_time(str_time):
    '''
    str类型的时间转换成time
    '''
    return datetime.datetime.strptime(str_time,TIME_FORMAT)
           
def get_agent_error_log(error_log_name, start_time):
    '''
    Failed to start XCTestAgent情况下获取xctest日志中的关键错误信息
    
    :param error_log_name 日志文件名
    :type str
    :param start_time 启动agent日志开始时间
    :type str 时间戳
    :return agent_error_log 返回agent出错的信息
    :type str or ''
    '''
    
    key_word = list(set(USER_KEY_WORDS).union(set(DEFAULT_KEY_WORDS)))
    error_pattern = get_match_pattern(key_word)
    
    time_pattern = r'^\[(\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2})'
    flag = False #标记该时间段的日志是否已经取完
    start_time = str_to_time(time.strftime(TIME_FORMAT,time.localtime(float(start_time))))
    end_time = str_to_time(time.strftime(TIME_FORMAT, time.localtime()))
    files = os.listdir(TMP_DIR_PATH)
    
    for f in files:
        log_path = os.path.join(TMP_DIR_PATH, f)
        if not flag and os.path.exists(log_path) and f.find(error_log_name) != -1:
            with open(log_path, "rb") as fd:
                for line in fd:
                    line = line.decode("utf-8")
                    match = re.search(time_pattern, line)
                    if match:
                        cur_time = str_to_time(match.group(1))
                        if cur_time < start_time:
                            flag = True
                            continue
                        elif cur_time > end_time:
                            break
                        else:
                            if re.findall(error_pattern, line):
                                return line
    
    return 'No matches found, please check the \'QT4I_AGENT_ERROR_KEY_WORDS\' configuration item'
    
def get_match_pattern(key_word):
    '''
    获取setting文件中配置的对应内容（一个元组），生成正则表达式匹配形式
    :param error_logo setting文件中配置的标志名
    :type str
    :return pattern
    :type str
    '''
    pattern = ''
    
    for key in key_word:
        pattern += '.*' + key + '.*|'
    return pattern[0:len(pattern)-1]



# -*- coding:utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''iOS基础UI控件模块
'''

from __future__ import absolute_import, print_function, division

import base64
import os
import re
import time
import traceback
import uuid
import hashlib
import six
from six import string_types
from six import StringIO
from six.moves.xmlrpc_client import Fault

from testbase.conf import settings
from testbase.util import LazyInit
from qt4i.exceptions import ControlAmbiguousError
from qt4i.exceptions import ControlNotFoundError
from qt4i.app import App
from qt4i.device import QT4i_LOGS_PATH
from qt4i.qpath import QPath
from qt4i.util import Rectangle, EnumDirect, Timeout


INS_IOS_DRIVER = True if settings.get('IOS_DRIVER', 'xctest') == 'instruments' else False
QTA_AI_SWITCH = settings.get('QT4I_AI_SWITCH', False)


class ControlContainer(object):
    '''控件集合接口
    '''
    
    def __init__(self):
        self._locators = {}  # 对象定义
        
    def __getitem__(self, key):
        '''操作符"[]"重载
        
        :param key: 控件名
        :type key: str
        :rtype: object
        :raises: Exception
        '''
        if not isinstance(key, string_types) or key not in self._locators: raise Exception('子控件名错误: %s' % key)
        params = self._locators.get(key)
        if isinstance(params, dict):
            cls = params.get('type')
            root = params.get('root')
            locator = params.get('locator')
            title = params.get('title')
            url = params.get('url')
            usr_name = params.get('name')
            mt_instance = params.get('instance')
            params = {'root': root, 'locator': locator}
            if INS_IOS_DRIVER:
                while True:
                    if not isinstance(root, string_types): break
                    if re.match('^@', root):
                        root_key = re.sub('^@', '', root)
                        if root_key not in self._locators:
                            raise ControlNotFoundError('未找到父控件: %s' % root_key)
                        root_params = self._locators.get(root_key)
                        params = {'root'    : root_params.get('root'),
                                  'locator' : str(root_params.get('locator')) + str(params.get('locator'))}
                        root = root_params.get('root')                
            else:
                if isinstance(root, string_types) and re.match('^@', root):
                    root_key = root[1:]
                    if root_key not in self._locators:
                        raise ControlNotFoundError('未找到父控件: %s' % root_key)
                    params['root'] = self[root_key]
            if usr_name:
                params['name'] = usr_name 
            params['id'] = key
            if title:
                params['title'] = title
            if url:
                params['url'] = url
            if mt_instance is not None:
                params['instance'] = mt_instance
            return cls(**params)
        raise Exception('控件定义的结构异常: [%s]' % key)
    
    @property
    def Controls(self):
        '''返回子控件集合 - 示例：XX_Window.Controls['XX_按钮'].click()
        '''
        return self
    
    def clearLocator(self):
        '''清空控件定位参数
        '''
        self._locators.clear()
        
    def hasControlKey(self, control_key):
        '''是否包含控件control_key
        
        :param control_key: 控件名
        :type control_key: str
        :rtype: boolean
        '''
        return control_key in self._locators
    
    def updateLocator(self, locators):
        '''更新控件定位参数
        
        :param locators: 定位参数，格式是 {'控件名':{'type':控件类, 控件类的参数dict列表}, ...}
        :type locators: dict
        '''
        self._locators.update(locators)

    # 此接口要被废弃，请使用 Element(...).wait_for_exist(10)
    def isChildCtrlExist(self, childctrlname, timeout=Timeout(2, 0.005)):
        '''判断子控件是否存在
        
        :param childctrlname: 控件名
        :type childctrlname: str
        :rtype: boolean
        '''
        print('*' * 30)
        print('提示: isChildCtrlExist 接口已过时，请使用 Element(...).wait_for_exist()')
        print('*' * 30)
        old_timeout = Element.timeout
        Element.timeout = timeout
        result = self.Controls[childctrlname].exist()
        Element.timeout = old_timeout
        return result


class _Element(object):

    def __init__(self, _id):
        self._id = _id

    @property
    def id(self):
        return self._id


class Element(ControlContainer):
    '''控件（UI框架是由各种控件组合而成）
    '''
    
    timeout = Timeout(5, 0.005)
    
    def __init__(self, root, locator, **ext):
        '''构造函数
        
        :param root: 父对象(多态：最顶层是App实例对象)
        :type root: App | Element
        :param locator: 定位描述（多态：可以直接是element的id），如果为None则指向_root对象
        :type locator: QPath | str(QPath) | int已定位到的element的ID
        '''
        ControlContainer.__init__(self)
        if ext:
            self.user_name = ext['id']
        self._app = None
        self._root = root
        self._locator = locator
        # -*- -*- -*-
        if not isinstance(self._root, (App, Element)):
            raise Exception('element.root is invalid')
        if self._locator is not None and not isinstance(self._locator, (QPath, string_types, int)):
            raise Exception('element.locator is invalid')
        # -*- -*- -*-
        root = self._root
        while True:
            if isinstance(root, App):
                self._app = root
                break
            root = root._root
        if not isinstance(self._app, App):
            raise Exception('element.app is invalid')
        # -*- -*- -*-
        self._check_locator(self._locator)
        # -*- -*- -*-
        self._element = LazyInit(self, '_element', self._init_element)
   
    def _check_locator(self, _locator):
        if isinstance(_locator, string_types) and not _locator.startswith('/'): #默认为_locator为id
            self.strategy = 'id'
        else:
            self.strategy = 'qpath'

    def _find(self, parent_id=None):
        print('--- ' * 9)
        print('locator : %s' % self._locator)
        print('timeout : %s' % self.timeout.timeout)
        # --- --- --- --- --- --- --- --- ---
        result = self._app.driver.element.find_elements((str(self._locator) if self._locator else None), self.timeout.timeout, self.timeout.interval, self.strategy, parent_id)
        elements = result.get('elements', [])
        element = elements[0].get('element') if len(elements)>0 else None
        # --- --- --- --- --- --- --- --- ---
        if element is None:
            if self.strategy == "id":
                err_msg  = '\n未找到控件 : 耗时[%s]毫秒尝试了[%s]次查找 [%s]' % (result.get('find_time'), result.get('find_count'), self._locator)
                err_msg += '\n无效控件ID : %s' % self._locator
                err_msg += '\nControlNotFoundError 建议用 device.print_uitree() 重新分析UI树'
            else:                
                err_msg  = '\n未找到控件 : 耗时[%s]毫秒尝试了[%s]次查找 [%s]' % (result.get('find_time'), result.get('find_count'), self._locator)
                err_msg += '\n有效Path  : %s' % result.get('valid_path_part')
                err_msg += '\n无效Path  : %s' % result.get('invalid_path_part')
                err_msg += '\nControlNotFoundError 建议用 device.print_uitree() 重新分析UI树'
            raise ControlNotFoundError(err_msg)
        if len(elements) > 1 and not INS_IOS_DRIVER:
            err_msg = '\n找到多个控件: 耗时[%s]毫秒尝试了[%s]次查找，找到[%s]个控件' % (result.get('find_time'), result.get('find_count'), len(elements))
            for e in elements: err_msg += '\n  %s' % e
            err_msg += '\nControlAmbiguousError 建议用 device.print_uitree() 重新分析UI树'
            raise ControlAmbiguousError(err_msg)
        print('element : 耗时[%s]毫秒尝试了[%s]次查找到[ element id: %s ]' % (result.get('find_time'), result.get('find_count'), element))
        return element

    def _init_element(self):
        if isinstance(self._locator, int): _id = self._locator
        else:
            if isinstance(self._root, Element) : _id = self._find(self._root._element.id)
            else                               : _id = self._find()
        if not isinstance(_id, int): raise Exception('element is invalid')
        self._element = _Element(_id)
        return self._element
    

    def exist(self):
        '''控件是否存在
        
        :rtype: boolean
        '''
        try:
            return bool(self._init_element())
        except Fault as err:
            raise err
        except Exception as err:
            print(err)
        return False

    def wait_for_exist(self, timeout, interval):
        '''控件是否存在
        
        :param timeout: 超时值（秒），例如对象需要等待网络拉取10秒才出现
        :type timeout: float
        :param interval: 轮询值（秒），轮询值建议 0.005 秒
        :type interval: float
        :rtype: boolean
        '''
        self.timeout = Timeout(timeout, interval)
        return self.exist()
    
    def find_elements(self, locator):
        '''通过相对于当前父对象的Path搜索子孙控件（建议用此方法解决动态Path的场景）
        
        :param locator: 相对当前对象的 - 子Path对象 或 子Path字符串 （搜索的起点为当前父对象）
        :type locator: QPath | str
        :return: [Element, ...]
        :rtype: list
        '''
        if not isinstance(locator, (QPath, string_types)): raise Exception('find_elements(path) path is invalid')
        strategy = 'qpath'
        if isinstance(locator, string_types) and not locator.startswith('/'):
            strategy = 'id'
        elements = []
        result = self._app.driver.element.find_elements(str(locator), self.timeout.timeout, self.timeout.interval, strategy, self._element.id)
        found_elements = result.get('elements', [])
        for item in found_elements: elements.append(Element(root=self, locator=item.get('element')))
        return elements

    def first_with_name(self, name):
        '''通过name文本获取第一个匹配的子element
        
        :param name: 子控件的name
        :type name: str
        :rtype: Element or None
        '''
        _id = self._app.driver.element.find_element_with_value_for_key(self._element.id, 'name', name)
        if isinstance(_id, int): return Element(root=self, locator=_id)

    def with_name(self, name):
        '''通过name文本获取匹配的子elements
        
        :param name: 子控件的name
        :type name: str
        :rtype: list
        '''
        children = []
        children_id = self._app.driver.element.find_elements_with_value_for_key(self._element.id, 'name', name)
        for child_id in children_id: children.append(Element(root=self, locator=child_id))
        return children

    def first_with_predicate(self, predicate):
        '''通过predicate文本获取第一个匹配的子element
        
        :param predicate: 预期子element的predicate （例如：“name beginswith 'xxx'”）
        :type predicate: str
        :rtype: Element or None
        '''
        _id = self._app.driver.element.find_element_with_predicate(self._element.id, predicate)
        if isinstance(_id, int): return Element(root=self, locator=_id)

    def with_predicate(self, predicate):
        '''通过predicate文本获取第一个匹配的子element
        :param predicate: 预期子element的predicate （例如：“name beginswith 'xxx'”）
        :type predicate: str
        :rtype: Element or None
        '''
        children = []
        children_id = self._app.driver.element.find_elements_with_predicate(self._element.id, predicate)
        for child_id in children_id: children.append(Element(root=self, locator=child_id))
        return children

    def first_with_value_for_key(self, key, value):
        '''通过匹配指定key的value，获取第一个匹配的子element

        :param key: key （例如：label、name、value）
        :type key: str
        :param value: 对应key的value值
        :type value: str or int
        :rtype: Element or None
        '''
        _id = self._app.driver.element.find_element_with_value_for_key(self._element.id, key, value)
        if isinstance(_id, int): return Element(root=self, locator=_id)

    def with_value_for_key(self, key, value):
        '''通过匹配指定key的value，获取所有匹配的子element
        
        :param key: key （例如：label、name、value）
        :type key: str
        :param value: 对应key的value值
        :type value: str or int
        :rtype: Element or None
        '''
        children = []
        children_id = self._app.driver.element.find_elements_with_value_for_key(self._element.id, key, value)
        for child_id in children_id: children.append(Element(root=self, locator=child_id))
        return children

    @property
    def children(self):
        '''获取子控件（仅含相对本控件的第二层子控件，不含第三层以及更深层的子控件，建议用此方法解决动态Path的场景）
        
        :return: [Element, ...]
        :rtype: list
        '''
        children = []
        children_id = self._app.driver.element.get_children_elements(self._element.id)
        for child_id in children_id:
            children.append(Element(root=self, locator=child_id))
        return children
    
    @property
    def parent(self):
        '''获取上一级的父控件
        
        :rtype: Element
        '''
        _id = self._app.driver.element.get_parent_element(self._element.id)
        if isinstance(_id, int): return Element(root=self, locator=_id)      
    
    def _get_attr(self, attr_name):
        '''获取控件的属性
        '''
        return self._app.driver.element.get_element_attr(self._element.id, attr_name)
    
    def get_attr_dict(self):
        '''获取元素的属性信息，返回字典
        
        :rtype: dict
        '''
        return self._app.driver.element.get_element_attrs(self._element.id)
    
    @property  
    def label(self):
        '''控件的label
        
        :rtype: str or None
        '''
        return self._get_attr('label')
    
    @property  
    def name(self):
        '''控件的name
        
        :rtype: str or None
        '''
        return self._get_attr('name')
    
    @property  
    def value(self):
        '''控件的value
        
        :rtype: None | str
        '''
        return self._get_attr('value')
    
    @value.setter
    def value(self, value):
        '''设置value(输入，支持中文)
        '''
        self._app.driver.element.set_value(self._element.id, value)
    
    @property  
    def visible(self):  
        '''控件是否可见
        
        :rtype: boolean
        '''
        return self._get_attr('visible')
    
    @property  
    def enabled(self): 
        '''控件是否开启
        
        :rtype: boolean
        ''' 
        return self._get_attr('enabled')
    
    @property
    def rect(self):
        '''控件的矩形信息
        
        :rtype: Rectangle
        '''
        rect = self._app.driver.element.get_rect(self._element.id)
        origin = rect['origin']
        size = rect['size']
        return Rectangle(origin['x'], origin['y'], origin['x'] + size['width'], origin['y'] + size['height'])

    def send_keys(self, keys):
        '''输入字符串

        :param keys: 字符串内容
        :type keys: str        
        :attention: 该接口不支持中文，中文输入请使用value='中文'
        '''
        self._app.driver.element.send_keys(self._element.id, keys)
    
    def scroll_to_visible(self, rate=1.0, drag_times=20):
        '''自动滚动到元素可见（技巧: Path中不写visible=true，当对象在屏幕可视范围之外，例如底部，调用此方法可以自动滚动到该元素为可见）
        
        :param drag_times: 最多尝试下滑次数，默认20次
        :type drag_times: int
        '''
        try:
            self._app.driver.element.scroll_to_visible(self._element.id, rate)
        except Exception:
                err = traceback.format_exc()
                if 'has no scrollable ancestor' not in err:
                    raise
        if not self.visible:  #兼容UIAutomation框架中scrollToVisible失效的场景
            for _ in range(drag_times):
                rate = self.rect.top / self._app.device.rect.height
                if rate > 1:
                    self._app.driver.device.drag(0.5, 0.6, 0.5 ,0.4)
                elif rate < -1:
                    self._app.driver.device.drag(0.5, 0.4, 0.5 ,0.6)
                elif rate >= -1 and rate < 0.5: 
                    self._app.driver.device.drag(0.5, 0.45, 0.5 ,0.55)
                elif rate >= 0.5 and rate <= 1: 
                    self._app.driver.device.drag(0.5, 0.55, 0.5 ,0.45)
                time.sleep(0.3)
                if self.visible:
                    break
            else:
                raise Exception('该控件不可见，请检查控件的位置[%s]和可见属性是否正常' % self.rect)
    
    def click(self, offset_x=None, offset_y=None):
        '''点击控件
        
        :param offset_x: 相对于该控件的坐标offset_x，百分比( 0 -> 1 )，不传入则默认该控件的中央
        :type offset_x: float or None
        :param offset_y: 相对于该控件的坐标offset_y，百分比( 0 -> 1 )，不传入则默认该控件的中央
        :type offset_y: float or None
        '''
        self._app.driver.element.click(self._element.id, offset_x, offset_y)
    
    def double_click(self, offset_x=0.5, offset_y=0.5):
        '''双击控件
        
        :param offset_x: 相对于该控件的坐标offset_x，百分比( 0 -> 1 )，不传入则默认该控件的中央
        :type offset_x: float
        :param offset_y: 相对于该控件的坐标offset_y，百分比( 0 -> 1 )，不传入则默认该控件的中央
        :type offset_y: float
        '''
        self._app.driver.element.double_click(self._element.id, offset_x, offset_y)
    
    def long_click(self, duration=3, offset_x=0.5, offset_y=0.5):
        '''单指长按
        
        :param duration: 持续时间（秒）
        :type duration: int
        :param offset_x: 相对于该控件的坐标offset_x，百分比( 0 -> 1 )，不传入则默认该控件的中央
        :type offset_x: float
        :param offset_y: 相对于该控件的坐标offset_y，百分比( 0 -> 1 )，不传入则默认该控件的中央
        :type offset_y: float
        '''
        self._app.driver.element.long_click(self._element.id, offset_x, offset_y, duration)
        
    def _tap_with_options(self, options={}):
        '''自定义点击(默认单指点击一次控件的中央)
        
        :param options : 自定义项
        :type options  : dict : {
                       :           tapCount     : 1,                    // 轻触次数
                       :           touchCount   : 1,                    // 触摸点
                       :           duration     : 1,                    // 持续时间
                       :           tapOffset    : { x: 0.5, y: 0.5 }    // 轻触偏移百分比
                       :          }
        '''
        options = {'tapCount'   : options.get('tapCount'  , 1),
                   'touchCount' : options.get('touchCount', 1),
                   'duration'   : options.get('duration'  , 0.1),
                   'tapOffset'  : {
                                   'x' : options.get('tapOffset', {}).get('x', 0.5),
                                   'y' : options.get('tapOffset', {}).get('y', 0.5)
                                  }}
        self._app.driver.element.click(self._element.id, None, None, options)
    
    def drag(self, from_x=0.9, from_y=0.5, to_x=0.1, to_y=0.5, duration=0.5):
        '''回避控件边缘，在控件体内拖拽（默认在控件内从右向左拖拽）
        
        :param from_x: 起点 x偏移百分比（从左至右为0.0至1.0）
        :type from_x: float
        :param from_y: 起点 y偏移百分比（从上至下为0.0至1.0）
        :type from_y: float
        :param to_x: 终点 x偏移百分比（从左至右为0.0至1.0）
        :type to_x: float
        :param to_y: 终点 y偏移百分比（从上至下为0.0至1.0）
        :type to_y: float
        :param duration: 持续时间（秒）
        :type duration: float
        '''
        self._app.driver.element.drag(self._element.id, from_x, from_y, to_x, to_y, duration)
    
    def drag2(self, direct=EnumDirect.Left):
        '''回避边缘在控件体内拖拽
        
        :param direct: 拖拽的方向
        :type  direct: EnumDirect.Left|EnumDirect.Right|EnumDirect.Up|EnumDirect.Down
        '''
        if direct == EnumDirect.Left  : self._app.driver.element.drag(self._element.id, 0.5, 0.5, 0.1, 0.5, 0.5)
        if direct == EnumDirect.Right : self._app.driver.element.drag(self._element.id, 0.5, 0.5, 0.9, 0.5, 0.5)
        if direct == EnumDirect.Up    : self._app.driver.element.drag(self._element.id, 0.5, 0.5, 0.5, 0.1, 0.5)
        if direct == EnumDirect.Down  : self._app.driver.element.drag(self._element.id, 0.5, 0.5, 0.5, 0.9, 0.5)
    
    def flick(self, from_x=0.9, from_y=0.5, to_x=0.1, to_y=0.5):
        '''滑动/拂去（默认从右向左回避边缘进行滑动/拂去）该接口比drag的滑动速度快，如果滚动距离大，建议用此接口
        
        :param from_x: 起点 x偏移百分比（从左至右为0.0至1.0）
        :type from_x: float
        :param from_y: 起点 y偏移百分比（从上至下为0.0至1.0）
        :type from_y: float
        :param to_x: 终点 x偏移百分比（从左至右为0.0至1.0）
        :type to_x: float
        :param to_y: 终点 y偏移百分比（从上至下为0.0至1.0）
        :type to_y: float
        '''
        self._app.driver.element.drag(self._element.id, from_x, from_y, to_x, to_y, 0)
    
    def flick2(self, direct=EnumDirect.Left):
        '''回避边缘在控件体内滑动/拂去
        
        :param direct: 滑动/拂去的方向
        :type  direct: EnumDirect.Left|EnumDirect.Right|EnumDirect.Up|EnumDirect.Down
        '''
        if direct == EnumDirect.Left  : self._app.driver.element.drag(self._element.id, 0.5, 0.5, 0.1, 0.5, 0)
        if direct == EnumDirect.Right : self._app.driver.element.drag(self._element.id, 0.5, 0.5, 0.9, 0.5, 0)
        if direct == EnumDirect.Up    : self._app.driver.element.drag(self._element.id, 0.5, 0.5, 0.5, 0.1, 0)
        if direct == EnumDirect.Down  : self._app.driver.element.drag(self._element.id, 0.5, 0.5, 0.5, 0.9, 0)
    
    def screenshot(self, image_path=None, image_type=None):
        '''截屏
        
        :attention: 裁剪图像使用了PIL库，使用该接口请安装Pillow，如下：pip install Pillow
        :param image_path: 截屏图片的存放路径
        :type image_path: str
        :param image_path: 截屏图片的存放路径
        :type image_path: str
        :return: 截屏结果和截屏文件的路径
        :rtype: tuple (boolean, str)
        '''
        base64_img = self._app.driver.device.capture_screen()
        device_img_data = base64.decodestring(base64_img)
        from PIL import Image
        image_fd = StringIO(device_img_data)
        img = Image.open(image_fd)
        scale_x = self._app.driver.device.get_rect()['size']['width']*1.0/img.size[0]
        scale_y = self._app.driver.device.get_rect()['size']['height']*1.0/img.size[1]
        rect = self.rect
        rect = (rect.left/scale_x, rect.top/scale_y, rect.right/scale_x, rect.bottom/scale_y)
        import math
        rect = (math.floor(rect[0]), math.floor(rect[1]), math.floor(rect[2]), math.floor(rect[3]))
        element_img = img.crop(rect)
        if image_type == 'Pil.Image':
            return element_img
        if not image_path:
            image_path = os.path.join(QT4i_LOGS_PATH, "p%s_%s.png" %(os.getpid(), uuid.uuid1()))
        element_img.save(os.path.abspath(image_path))
        return (os.path.isfile(image_path), image_path)

    def print_uitree(self):
        '''打印当前控件的UI树
        '''
        _ui_tree = self._app.driver.element.get_element_tree(self._element.id)
        def _print(_tree, _spaces='', _indent='|---'):
            _line = _spaces + '{  ' + ',   '.join([
                'classname: "%s"' % _tree['classname'],
                'label: %s' % ('"%s"' % _tree['label'] if _tree['label'] else 'null'),
                'name: %s' % ('"%s"' % _tree['name' ] if _tree['name' ] else 'null'),
                'value: %s' % ('"%s"' % _tree['value'] if _tree['value'] else 'null'),
                'visible: %s' % ('true' if _tree['visible'] else 'false'),
                'enabled: %s' % ('true' if _tree['enabled'] else 'false'),
                'rect: %s' % _tree['rect']]) + '  }'
            print(_line)
            for _child in _tree['children']: _print(_child, _spaces + _indent, _indent)
        _print(_ui_tree)
   
    def get_metis_view(self):
        '''返回MetisView
        '''
        return MetisView(self)
    
    def force_touch(self, pressure=1.0, duration=2.0):
        '''3D touch

        :param pressure: 按压力度（取值：0-1.0）
        :type pressure: float
        :param duration: 按压持续的时间（单位:秒）
        :type duration: float
        '''
        self._app.driver.element.force_touch(self._element.id, pressure, duration)


class Window(Element):
    '''窗口基类
    '''
    def __init__(self, root, locator=None, **ext):
        Element.__init__(self, root, (locator if locator else 1), **ext)
        

class Alert(Element):
    '''弹出框
    
    由于弹出框在iOS不同版本中的UI逻辑不一致（QPath不同），但业务逻辑相对统一（确定/取消），故单独封装成一个类
    '''
    
    def __init__(self, root, locator= QPath("/classname='UIAAlert' & maxdepth = 20"), **ext):
        Element.__init__(self, root=root, locator=locator, **ext)
        self._ios_version = self._app.device.ios_version
        self._title = None
        self._buttons = []

    @property
    def title(self):
        '''提示文本内容
        '''
        if INS_IOS_DRIVER:
            self._title = Element(root=self, locator=QPath("/classname='UIAScrollView'/classname='UIAStaticText' && visible=true"))
        else:
            self._title = Element(root=self, locator=QPath("/classname='StaticText' & visible=true & maxdepth=10"))
            
        if self._title is not None: 
            return self._title
        raise Exception('未找到有效标题，请检查Alert类内的子对象是否正确的QPath')

    @property
    def buttons(self):
        '''Alert控件上的所有按钮
        '''
        if INS_IOS_DRIVER:
            if re.search('^[8|9]', self._ios_version):
                self._buttons = self.find_elements(QPath("/classname='UIACollectionView'/classname='UIACollectionCell'/classname='UIAButton' && visible=true"))
            elif re.search('^7', self._ios_version):
                self._buttons = self.find_elements(QPath("/classname='UIATableView'/classname='UIATableCell' && visible=true"))
            if len(self._buttons) == 0: 
                self._buttons = self.find_elements(QPath("/classname='UIACollectionView'/classname='UIACollectionCell' && visible=true"))
        else:
            self._buttons = self.find_elements(QPath("/classname='Button' & visible=true & maxdepth=10"))
        if len(self._buttons) > 0                           : return self._buttons
        raise Exception('未找到有效按钮，请检查Alert类内的子对象是否正确的QPath')

    def click_button(self, text):
        '''点击指定文本的按钮
        
        :param text: 按钮上的文本
        :type text: str
        '''
        for button in self.buttons:
            text = str(text)
            attr = button.get_attr_dict()
            label = str(attr['label'])
            name = str(attr['name'])
            value = str(attr['value'])
            if text in [label, name, value]:
                button.click()
                return
        raise Exception('未有效匹配按钮，请检查指定的字符串是否命中预期')

    
class Slider(Element):
    '''进度条
    '''
    
    def __init__(self, root, locator=QPath("/classname='UIASlider' & maxdepth = 20"), **ext):
        Element.__init__(self, root=root, locator=locator, **ext)
        self._ios_version = self._app.device.ios_version
    
    @Window.value.setter
    def value(self, value):
        '''设置进度条的值(取值0~1之间)
        '''
        self._app.driver.element.drag_to_value(self._element.id, value)

        
class ActionSheet(Element): 
    '''UIActionSheet弹出框
    '''
    
    def __init__(self, root, locator=None, **ext):
        if locator is None:
            if INS_IOS_DRIVER:
                locator =  QPath("/classname='UIAWindow'/classname='UIAActionSheet'")    
            else:
                locator =  QPath("/classname='Sheet' & maxdepth = 20") 
        Element.__init__(self, root=root, locator=locator, **ext)
        self._ios_version = self._app.device.ios_version  
        self._buttons = []
    
    @property
    def buttons(self):  
        '''ActionSheet上面的所有按钮
        '''
        if INS_IOS_DRIVER:
            if re.search('^7', self._ios_version):
                self._buttons = self.find_elements(QPath("/classname='UIAButton' && visible=true"))
            else:
                self._buttons = self.find_elements(QPath("/classname='UIACollectionView'/classname='UIACollectionCell'/classname='UIAButton' && visible=true"))
                other_buttons = self.find_elements(QPath("/classname='UIAButton' && visible=true"))
                if other_buttons:
                    self._buttons.extend(other_buttons)
        else:
            self._buttons = self.find_elements(QPath("/classname='Button' & visible=true & maxdepth=10"))
        
        if len(self._buttons) > 0:
            return self._buttons
        raise Exception('未找到有效按钮，请检查ActionSheet类内的控件QPath是否正确')
    
    def click_button(self, text):
        '''点击text对应的按钮
        
        :param text: 按钮上的文本
        :type text: str
        '''
        for button in self.buttons:
            text = str(text)
            attr = button.get_attr_dict()
            label = str(attr['label'])
            name = str(attr['name'])
            value = str(attr['value'])
            if text in [label, name, value]:
                button.click()
                return
        raise Exception('未有效匹配按钮，请检查指定的字符串是否命中预期')            


class TableView(Element):
    '''TableView控件
    '''
    
    def __init__(self, root, locator=None, **ext):
        if locator is None:
            if INS_IOS_DRIVER:
                locator =  QPath("/classname='UIAWindow'/classname='UIATableView'")    
            else:
                locator =  QPath("/classname='Table' & maxdepth = 20")
        Element.__init__(self, root, locator, **ext)
        if isinstance(root, Window):
            for key, value in six.iteritems(root._locators):
                if value['root'] == '@%s' % self.user_name:
                    self.updateLocator({key:value})
        self._cells = []
        
    
    @property
    def cells(self):
        if INS_IOS_DRIVER:
            self._cells = self.find_elements("/classname='UIATableCell'")
        else:
            self._cells = self.find_elements("/classname='Cell'")
        
        if len(self._cells) > 0:
            return self._cells      
        raise RuntimeError('未找到有效TableCell，请检查TableView类内的控件QPath是否正确')
    
    def click_cell(self, text):
        for cell in self.cells:
            if text in cell.name: # 考虑效率问题，目前只查找name
                if not cell.visible:
                    cell.scroll_to_visible()
                cell.click()
                return 
        raise RuntimeError('未找到指定TableCell:%s' % text)
    
    def __iter__(self):
        for cell in self.cells:
            yield Cell(cell, self._locators)


class Cell(object):
    '''TableCell控件
    '''
    def __init__(self, element, locators={}):
        self._element = element
        for key in locators:
            locators[key]['root'] = self._element
        self._element.updateLocator(locators)
        
    def __getattr__(self, attr):
        return getattr(self._element, attr) 
    

class Button(Element):
    '''Button控件
    '''

    
class TextField(Element):
    '''文本输入框
    '''
    
    
class SecureTextField(Element):
    '''密码输入框
    '''
    
    
class PickerWheel(Element):
    '''滚轮选择框
    '''
    
    def select(self, value):
        self._app.driver.element.select_picker_wheel(self._element.id, value)

        
class MetisView(object):
    '''MetisView
    '''
    def __init__(self, element):
        self.element = element
        if isinstance(self.element, Element):
            img = self.element.screenshot(image_type='Pil.Image')
            real_width = img.size[0]
            real_height = img.size[1]
            rect = self.element.rect
            self._dx = real_width*1.0/rect.width
            self._dy = real_height*1.0/rect.height

    @property
    def rect(self):
        '''元素相对坐标(x, y, w, h)
        '''
        rect = self.element.rect
        if isinstance(self.element, Element):
            return (rect.top*self._dy, rect.left*self._dx, rect.width*self._dx, rect.height*self._dy)
        else:
            return (rect.top, rect.left, rect.width, rect.height)

    @property
    def os_type(self):
        '''系统类型，例如"android"，"ios"，"pc"
        '''
        return "ios"

    def screenshot(self):
        '''当前容器的区域截图
        '''
        return self.element.screenshot(image_type='Pil.Image')

    def click(self, offset_x=None, offset_y=None):
        '''点击
        
        :param offset_x: 相对于该控件的坐标offset_x，百分比( 0 -> 1 )，不传入则默认该控件的中央
        :type offset_x: float or None
        :param offset_y: 相对于该控件的坐标offset_y，百分比( 0 -> 1 )，不传入则默认该控件的中央
        :type offset_y: float or None
        '''
        self.element.click(offset_x, offset_y)

    def send_keys(self, text):
        self.element.send_keys(text)

    def double_click(self, offset_x=None, offset_y=None):
        if offset_x is None:
            offset_x = 0.5
        if offset_y is None:
            offset_y = 0.5
        self.element.double_click(offset_x, offset_y)

    def long_click(self, duration=3, offset_x=None, offset_y=None):
        if offset_x is None:
            offset_x = 0.5
        if offset_y is None:
            offset_y = 0.5
        self.element.long_click(duration, offset_x=offset_x, offset_y=offset_y)
    
    def drag(self, from_x=0.5, from_y=0.5, to_x=0.5, to_y=0.1, duration=0.5):
        '''拖拽
        
        :param from_x: 起点 x偏移百分比（从左至右为0.0至1.0）
        :type from_x: float
        :param from_y: 起点 y偏移百分比（从上至下为0.0至1.0）
        :type from_y: float
        :param to_x: 终点 x偏移百分比（从左至右为0.0至1.0）
        :type to_x: float
        :param to_y: 终点 y偏移百分比（从上至下为0.0至1.0）
        :type to_y: float
        :param duration: 持续时间（秒）
        :type duration: float
        '''
        self.element.drag(from_x, from_y, to_x, to_y, duration)


# -*- coding:utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''远程控制的截屏服务
'''

from __future__ import absolute_import, print_function

import Queue
import threading
import time
import struct
import subprocess
import re
import os
import traceback

import signal
import argparse
import atexit
import sys
from six.moves.socketserver import BaseRequestHandler
from six.moves.socketserver import TCPServer
from six.moves.socketserver import ThreadingMixIn
from six.moves.urllib.request import urlopen


from qt4i.driver.tools import logger as logging
from qt4i.driver.util import Process


DEFAULT_IP = '0.0.0.0'
DEFAULT_PORT = 12306
DEFAULT_AGENT_PORT = 8100
DEFAULT_PID_FILE = '/tmp/screencapure.pid'

logger = logging.get_logger()


class ArgumentParser(argparse.ArgumentParser):
    '''Argument parser, automatically display help while error
    '''

    def error(self, message):
        sys.stderr.write('error: %s\n' % message)
        self.print_help()
        sys.exit(2)


CMD_HELP = {
    "start": "start screencapure service daemon",
    "stop": "stop screencapure service daemon",
    "restart": "stop screencapure service daemon and start it again",
    "debug": "try to start screencapure service as a normal process for debug",
}

# 参数解析器
main_parser = ArgumentParser(description='QT4i screencapure program',
                             formatter_class=argparse.RawDescriptionHelpFormatter)

main_parser.add_argument('--pidfile', '-p', metavar='PIDFILE', dest='pidfile', default=DEFAULT_PID_FILE,
                         help='the path of pid file, %s is used by default' % DEFAULT_PID_FILE)

main_parser.add_argument('--udid', '-u', metavar='UDID', dest='udid', default=None,
                         help='udid of device')

main_parser.add_argument('--listen_port', '-l', metavar='LISTENPORT', dest='listen_port', default=DEFAULT_PORT,
                         help='listening port of driver server, port %s is used by default' % DEFAULT_PORT)

main_parser.add_argument('--mjpeg_port', '-m', metavar='MJPEGPORT', dest='mjpeg_port', default=DEFAULT_AGENT_PORT,
                         help='listening port of xctest agent, port %s is used by default' % DEFAULT_AGENT_PORT)

subparser_dict = {}
subparsers = main_parser.add_subparsers(title='List of Commands', metavar='COMMAND')
for command in CMD_HELP:
    subparser_dict[command] = subparsers.add_parser(command, help=CMD_HELP[command])
    subparser_dict[command].set_defaults(func=command)

subparser_dict['stop'].add_argument('--force', '-f', action='store_true', default=False,
                                    help='force stop driver server')


class Daemon(object):
    """
    A generic daemon class.
    Usage: subclass the Daemon class and override the run() method
    """

    def __init__(self, pidfile, stdin='/dev/null', stdout='/dev/null', stderr='/dev/null'):
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.pidfile = pidfile

    def daemonize(self):
        """
        do the UNIX double-fork magic, see Stevens' "Advanced
        Programming in the UNIX Environment" for details (ISBN 0201563177)
        http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
        """
        # Do first fork
        self.fork()

        # Decouple from parent environment
        self.dettach_env()
        # Do second fork
        self.fork()

        # Flush standart file descriptors
        sys.stdout.flush()
        sys.stderr.flush()

        #
        self.attach_stream('stdin', mode='r')
        self.create_pidfile()

    def attach_stream(self, name, mode):
        """
        Replaces the stream with new one
        """
        stream = open(getattr(self, name), mode)
        os.dup2(stream.fileno(), getattr(sys, name).fileno())

    def dettach_env(self):
        os.chdir("/")
        os.setsid()
        os.umask(0)

    def fork(self):
        """
        Spawn the child process
        """
        try:
            pid = os.fork()
            if pid > 0:
                sys.exit(0)
        except OSError as e:
            sys.stderr.write("Fork failed: %d (%s)\n" % (e.errno, e.strerror))
            sys.exit(1)

    def create_pidfile(self):
        atexit.register(self.delpid)
        pid = str(os.getpid())
        open(self.pidfile, 'w+').write("%s\n" % pid)

    def delpid(self):
        """
        Removes the pidfile on process exit
        """
        os.remove(self.pidfile)

    def start(self, is_daemon=True):
        """
        Start the daemon
        """
        # Check for a pidfile to see if the daemon already runs
        pid = self.get_pid()

        if pid:
            if Process().get_process_by_pid(pid, 'screencapture.py'):
                message = "pidfile %s already exist. Daemon already running?\n"
                logger.warning(message % self.pidfile)
                return
            else:
                os.remove(self.pidfile)

        # Start the daemon
        if is_daemon:
            self.daemonize()
        self.run()

    def get_pid(self):
        """
        Returns the PID from pidfile
        """
        try:
            pf = open(self.pidfile, 'r')
            pid = int(pf.read().strip())
            pf.close()
        except (IOError, TypeError):
            pid = None
        return pid

    def stop(self, silent=False):
        """
        Stop the daemon
        """
        # Get the pid from the pidfile
        pid = self.get_pid()

        if not pid:
            if not silent:
                message = "pidfile %s does not exist. Daemon not running?\n"
                logger.error(message % self.pidfile)
            return

            # Try killing the daemon process
        try:
            while True:
                os.kill(pid, signal.SIGTERM)
                time.sleep(0.1)
        except OSError as err:
            err = str(err)
            if err.find("No such process") > 0:
                if os.path.exists(self.pidfile):
                    os.remove(self.pidfile)
            else:
                logger.error(str(err))
                sys.exit(1)

    def restart(self):
        """
        Restart the daemon
        """
        self.stop(silent=True)
        self.start()

    def run(self):
        """
        You should override this method when you subclass Daemon. It will be called after the process has been
        daemonized by start() or restart().
        """
        raise NotImplementedError


class ThreadedTCPServer(ThreadingMixIn, TCPServer):
    allow_reuse_address = True
    request_queue_size = 16


class ScreenConsumer(BaseRequestHandler):

    def setup(self):
        self.screenqueue = Queue.Queue(24)
        self.screen_producer = self.server.screen_producer
        assert self.request.recv(10) == 'screenshot'
        self.screen_producer.add_client(self)

    def handle(self):
        while True:
            image_data = self.screenqueue.get()
            try:
                image_length = len(image_data)
                self.request.sendall(struct.pack("I", image_length) + image_data)
            except:
                break

    def finish(self):
        self.request.close()
        self.screen_producer.remove_client(self)


class ScreenCaptureService(Daemon):

    def __init__(self, pidfile, udid, listen_port, mjpeg_port):
        Daemon.__init__(self, pidfile)
        self._udid = udid
        self._listen_port = listen_port
        self._mjpeg_port = mjpeg_port

    def start(self, is_daemon=True):
        '''启动service
        '''
        pid = self.get_pid()
        if pid:
            process = Process().get_process_by_pid(pid, 'screencapture.py')
            if process:
                Process().kill_process_by_pid(pid)
                logger.warning('Kill old screencapure server process %d' % pid)
            else:
                os.remove(self.pidfile)

        # start damon
        if is_daemon:
            self.daemonize()
        self.run()

    def run(self):
        if self._udid:
            if self._mjpeg_port:
                screen_producer = ScreenProducer(self._udid, int(self._mjpeg_port))
                screen_producer.start()
                server = ThreadedTCPServer(('localhost', int(self._listen_port)), ScreenConsumer)
                server.screen_producer = screen_producer
                server.serve_forever()

    def stop(self, silent=False):
        '''停止service
        '''
        Daemon.stop(self, silent)


class MJpegClient(object):

    def __init__(self, url):
        self.stream = urlopen(url)
        self.boundary = None

    def parse_mjpeg_boundary(self):
        '''
        parse mjpeg stream boundary
        :param stream: mjpeg client stream
        :type stream: http.client.HTTPResponse
        :returns: str -- boundary
        '''

        if self.stream.getcode() != 200:
            raise Exception('Invalid response from server: %d' % self.stream.status)
        h = self.stream.info()
        content_type = h.get('Content-Type', None)
        match = re.search(r'boundary="?(.*)"?', content_type)
        if match is None:
            raise Exception('Content-Type header does not provide boundary string')
        self.boundary = match.group(1)

    def read_mjpeg_frame(self):
        hdr = self._read_headers(self.boundary)

        clen = self._parse_content_length(hdr)
        if clen == 0:
            raise EOFError('End of stream reached')

        self._check_content_type(hdr, 'image/jpg')
        left = clen
        buf = ''
        while left > 0:
            tmp = self.stream.read(left)
            buf += tmp
            left -= len(tmp)
        return buf

    def close(self):
        self.stream.close()

    def _read_header_line(self):
        '''Read one header line within the stream.
        The headers come right after the boundary marker and usually contain
        headers like Content-Type and Content-Length which determine the type and
        length of the data portion.
        '''
        return self.stream.readline().strip()

    def _read_headers(self, boundary):
        '''Read and return stream headers.
        Each stream data packet starts with an empty line, followed by a boundary
        marker, followed by zero or more headers, followed by an empty line,
        followed by actual data. This function reads and parses the entire header
        section. It returns a dictionary with all the headers. Header names are
        converted to lower case. Each value in the dictionary is a list of header
        fields values.
        '''
        l = ''
        while True:
            l = self._read_header_line()
            if l != '':
                break
        if l != boundary:
            raise Exception('Boundary string expected, but not found')

        headers = {}
        while True:
            l = self._read_header_line()
            # An empty line indicates the end of the header section
            if l == '':
                break

            # Parse the header into lower case header name and header body
            i = l.find(':')
            if i == -1:
                raise Exception('Invalid header line: ' + l)
            name = l[:i].lower()
            body = l[i + 1:].strip()

            lst = headers.get(name, list())
            lst.append(body)
            headers[name] = lst

        return headers

    def _parse_content_length(self, headers):
        # Parse and check Content-Length. The header must be present in
        # each chunk, otherwise we wouldn't know how much data to read.
        clen = headers.get('content-length', None)
        try:
            return int(clen[0])
        except (ValueError, TypeError):
            raise Exception('Invalid or missing Content-Length')

    def _check_content_type(self, headers, expected_type):
        ctype = headers.get('content-type', None)
        if ctype is None:
            raise Exception('Missing Content-Type header')
        ctype = ctype[0]

        i = ctype.find(';')
        if i != -1:
            ctype = ctype[:i]
        if ctype != expected_type:
            raise Exception('Wrong Content-Type: %s' % ctype)
        return True


class ScreenProducer(threading.Thread):

    def __init__(self, udid, mjpeg_port):
        super(ScreenProducer, self).__init__()
        self._clients = set()
        self._udid = udid
        self.running = True
        self._mutex = threading.Lock()
        self.no_client = True
        self.mjpeg_conn_event = threading.Event()
        self.mjpeg_conn_event.clear()

        if re.match(r'^[a-f0-9]+$', udid):
            self._is_simulator = False
            self._url = "http://127.0.0.1:%s" % mjpeg_port
        else:
            self._is_simulator = True

    @property
    def is_sim(self):
        return self._is_simulator

    def add_client(self, client):
        with self._mutex:
            self._clients.add(client)
        self.no_client = False
        self.mjpeg_conn_event.set()

    def remove_client(self, client):
        with self._mutex:
            self._clients.discard(client)
            if len(self._clients) == 0:
                self.no_client = True
                self.mjpeg_conn_event.clear()

    def process_mjpeg_stream(self, mjpeg_client):
        mjpeg_client.parse_mjpeg_boundary()

        while True:
            jpeg_frame = mjpeg_client.read_mjpeg_frame()
            with self._mutex:
                if len(self._clients) == 0:
                    break
            for client in self._clients:
                if client.screenqueue.full():
                    continue
                client.screenqueue.put(jpeg_frame)

    def _capture_screen_by_xctestagent(self):
        try:
            if self.no_client:
                self.mjpeg_conn_event.wait()
            mjpeg_client = MJpegClient(self._url)
            self.process_mjpeg_stream(mjpeg_client)
        except EOFError:
            pass
        except Exception:
            logger.exception(traceback.format_exc())
        finally:
            mjpeg_client.close()

    def _capture_screen_by_simctl(self):
        try:
            if self.no_client:
                self.mjpeg_conn_event.wait()
            cmd = 'xcrun simctl io %s screenshot --type=jpeg -' % self._udid
            jpeg_frame = subprocess.check_output(cmd, shell=True)
            with self._mutex:
                if len(self._clients) == 0:
                    return
            for client in self._clients:
                if client.screenqueue.full():
                    continue
                client.screenqueue.put(jpeg_frame)
        except:
            logger.exception(traceback.format_exc())
            time.sleep(1)

    def run(self):
        while self.running:
            if self.is_sim:
                self._capture_screen_by_simctl()
            else:
                self._capture_screen_by_xctestagent()

    def stop(self):
        self.running = False


if __name__ == '__main__':
    args = main_parser.parse_args()
    sc = ScreenCaptureService(args.pidfile, args.udid, args.listen_port, args.mjpeg_port)

    if args.func == 'start':
        sc.start()

    elif args.func == 'stop':
        sc.stop()

    elif args.func == 'restart':
        sc.restart()

    elif args.func == 'debug':
        sc.start(False)

    else:
        main_parser.print_help()
        sys.exit(1)



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''Instruments Manager driver
'''

from __future__ import absolute_import, print_function

import os
import threading
import time 
import urllib

from qt4i.driver.rpc import rpc_method
from qt4i.driver.rpc import RPCEndpoint
from qt4i.driver.tools import logger

from qt4i.driver.instruments.internal._command import DeviceCommandDelegator
from qt4i.driver.instruments.internal._jsonp import Json
from qt4i.driver.util import Process
from qt4i.driver.util import Task
from qt4i.driver.util import ThreadTask

CurrentDirectoryPath = os.path.abspath(os.path.dirname(__file__))


class Instruments(object):
    '''Instrument API
    '''
    
    instances= {}  #保存Instruments的实例,key为设备的udid
    CommandTimeout = 90
    if os.path.exists("/tmp/instruments.log"):
        os.remove("/tmp/instruments.log")
    
    device_command_delegator = DeviceCommandDelegator()
    PythonPath = Task("which python").execute()
    InstrumentDriverPath = os.path.join(CurrentDirectoryPath,"internal","instruments.py")
    TasksPID = {}
    TasksLastError = {}
    TasksBoostrapStandBy = {}
    TasksEnvironment = {}
    TasksStartedEvents = {}
    TasksStoppedEvents = {}
    LastExecCommandDeviceUdid = None
    
    def __init__(self, device_id):
        self.log = logger.get_logger("instruments")
        self.udid = device_id
        if self.udid not in self.instances:
            self.instances[self.udid] = self
        self.crash_flag = False


    def notify(self, json_data):
        '''notify
        
        :param json_data: 要通知的内容
        :type json_data: Json
        '''
        dict_data = Json().loads(json_data)
        method = dict_data["method"]
        params = dict_data["params"]
        device_udid = params[0]["environment"]["device_udid"]
        instruments_pid = params[0]["pid"]
        
        if method == "InstrumentsStarted":
            self.log.info(method)
        if method in ['TraceTemplateInvalid', 'TargetAppBundleIDInvalid', 'DeviceIsCurrentlyLockedWithAPasscode',
               'TheAppMustBeSignedWithADevelopmentIdentity', 'TargetAppIsNotFrontmost', 'TargetAppDied',
               'DeviceDisabled', 'InstrumentsError', 'InstrumentsTimeout']:
            if method == "TargetAppDied":
                self.crash_flag = True
            self.device_command_delegator.clean_all(device_udid, method)
            self.TasksLastError.update({device_udid:method})
            self.dispatch_stopped_events()
            Process().kill_process_tree_by_pid(instruments_pid)
            self.log.error(method)
        
        if method == "InstrumentsTerminated":
            self.device_command_delegator.clean_all(device_udid, method)
            self.TasksEnvironment.pop(device_udid, None)
            self.TasksBoostrapStandBy.pop(device_udid, None)
            self.TasksPID.pop(device_udid, None)
            self.LastExecCommandDeviceUdid = None
            self.dispatch_stopped_events()
            self.log.info(method)
         
    def send_result(self, result):
        '''发送结果给指定的设备
        
        :param result: 要发送的结果
        :type result: str or dict
        '''
        result = Json().loads(result)
        self.device_command_delegator.set_command_result(self.udid, result)
    
    def send_result_and_get_next(self, result, timeout=CommandTimeout):
        '''发送结果给指定设备，并获取下一条指令
        
        :param result: 要发送的结果
        :type result: list or dict
        :param timeout: 指令执行超时 (秒)
        :type timeout: int
        :returns : Json -- 下一条执行的指令
        '''
        self.LastExecCommandDeviceUdid = self.udid
        result = Json().loads(result)
        if result.get('id') is None or result.get('id') < 0:
            if result.get('result') == "BootstrapStandBy":
                self.TasksBoostrapStandBy.update({self.udid:True})
                self.dispatch_started_events()
        else:
            self.device_command_delegator.set_command_result(self.udid, result)
        command = self.device_command_delegator.get_next_command(self.udid, timeout)
        if command:
            return Json().dumps(command, indent=None)
        
    def start(self, bundle_id, app_params, env, xmlrpc_uri, trace_template=None, trace_output=None, retry=3, timeout=55):
        '''用指定设备启动指定APP在后台运行
        
        :param bundle_id: 要启动的APP的Bundle ID
        :type bundle_id: str
        :param app_params: APP 使用的参数
        :type app_params: dict
        :param env: 任务设备的环境
        :type env: dict
        :param trace_template: 专项测试使用trace_template的路径
        :type trace_template: str
        :param trace_output: 专项测试使用trace_output
        :type trace_output: str
        :param retry: 启动失败重试次数
        :type retry: int
        :param timeout: 单次启动超时 (秒)
        :type timeout: int
        :returns : boolean -- True if start successfully, False if not
        ''' 
        if timeout < 10:
            raise Exception("timeout must be longer than 10 seconds")
        if self.is_working(): 
            self.release()
        if env == None:
            env = {}
        self.TasksEnvironment.update({self.udid:env})
        params = ''
        if app_params:
            params_list = []
            for k in app_params:
                params_list.append("-%s" % k)
                params_list.append("%s" % app_params[k])
            params = ' '.join(params_list)
        for i in range(retry):
            self.log.info("ins start app retry %d of %d times"%(i+1, retry))
            ins_py_cmd = ' '.join([self.PythonPath, '"%s"'%self.InstrumentDriverPath,
                                '-device_udid "%s"' % self.udid,
                                '-device_simulator %s' % False,
                                '-bundle_id "%s"' % bundle_id,
                                '-params "%s"' % params if params else '',
                                '-trace_template "%s"' % trace_template,
                                '-trace_output "%s"' % trace_output,
                                '-xmlrpc_uri "%s"' % xmlrpc_uri,
                                '-timeout %s' % (20*60),
                                '-cmd_fetch_delegate_timeout %s' % 10,
                                '-environment "%s"' % urllib.quote("%s"%env)
                                ])
            self.log.info(ins_py_cmd)
            task = ThreadTask(ins_py_cmd, stdout_line_callback=self.log.info, stderr_line_callback=self.log.error)
            if self.wait_for_start(timeout):
                self.TasksPID.update({self.udid: task.pid})
                self.log.info("ins start app successfully")
                return True
            Process().kill_child_processes_by_ppid(task.pid)
            time.sleep(3)
        error = self.TasksLastError.pop(self.udid,None)
        if error:
            raise Exception("[InstrumentsManager] %s"%error)
        return False
    
    def is_working(self):
        '''检查当前app是否在工作
        
        :returns : boolean True if it is working, False if not
        '''
#         return self.udid in self.instances
        return bool(self.TasksBoostrapStandBy.get(self.udid))
      
    def exec_command(self, command, timeout = CommandTimeout):
        '''执行指令
        
        :param command: 指令，包含方法和参数
        :type command: json 例如 {"method": func, "params": list(params)}
        :param timeout: 指令执行超时等待时常
        :type timeout: int 
        :returns : 执行结果 
        '''
        if not self.is_working():
            raise Exception("[Instruments] The app is not working")
        result = self.device_command_delegator.exec_command(self.udid, command, timeout)
        error = result.get("error")
        if error:
            raise Exception(error)
        return result.get("result")

    def dispatch_started_events(self):
        '''去除已经开始的事件
        '''
        events = self.TasksStartedEvents.pop(self.udid, [])
        for event in events:
            event.set()
            events.remove(event)
            
    def dispatch_stopped_events(self):
        '''去除已经停止的事件
        '''
        events = self.TasksStoppedEvents.pop(self.udid, [])
        for event in events:
            event.set()
            events.remove(event)
            
    def wait_for_start(self, timeout = CommandTimeout):
        '''等待设备开始
        
        :param timeout: 等待超时 (秒)
        :type timeout: int
        :returns : boolean -- True if start successfully, False if not
        '''
        event = threading.Event()
        if self.udid in self.TasksStartedEvents:
            self.TasksStartedEvents[self.udid].append(event)
        else:
            self.TasksStartedEvents[self.udid] = [event]
        event.wait(timeout)
        return self.is_working()
    
    def wait_for_stop(self, timeout=CommandTimeout):
        '''等待设备停止
        
        :param timeout: 执行超时时长
        :type timeout: int
        :returns : boolean -- True if stop successfully, False if not
        '''
        event = threading.Event()
        if self.udid in self.TasksStoppedEvents:
            self.TasksStoppedEvents[self.udid].append(event)
        else:
            self.TasksStoppedEvents[self.udid] = [event]
        event.wait(timeout)
        return self.is_working() == False
    
    def release(self, timeout=CommandTimeout):
        '''释放某个设备的instrument
        
        :param timeout: 等待超时 (秒)
        :type timeout: int
        :returns : boolean -- True if release successfully, False if not
        '''
        _begin_time = time.time()
        if self.is_working():
            self.device_command_delegator.clean_all(self.udid, "InstrumentsRelease")
            self.device_command_delegator.send_command(self.udid, {"method": "release"}, timeout-(time.time()-_begin_time))
            if not self.wait_for_stop(timeout-(time.time()-_begin_time)):
                Process().kill_child_processes_by_ppid(self.TasksPID.get(self.udid))
            if self.is_working():
                Process().kill_process_by_name("_fetch_cmd_delegate.py")
                Process().kill_process_by_name("instruments.py")
                os.system('killall -9 "instruments" "Instruments"')
                self.TasksBoostrapStandBy.pop(self.udid, None)
                self.instances.pop(self.udid)
        time.sleep(3)
        return self.is_working() == False
    
    @classmethod
    def release_all(cls):
        '''释放当前使用的所有设备的instrument
        '''
        for device_udid in cls.instances:
            cls.instances(device_udid).release()
        
    def get_crash_flag(self):
        '''获取app是否crash的标识
        
        :returns : boolean -- True if app crashed, False if not
        '''
        return self.crash_flag
    
    def set_command_timeout(self, timeout=CommandTimeout):
        '''设置指令执行超时
        
        :param timeout: 指令执行超时
        :type timeout: int
        '''
        self.__class__.CommandTimeout = timeout

    def get_driver_log(self):
        driver_log_path = None
        for h in self.log.handlers:
            attr_name = 'baseFilename'
            if hasattr(h, attr_name):
                driver_log_path = getattr(h, attr_name)
                break
        return driver_log_path

class InstrumentsCallBacker(RPCEndpoint):
    '''Instruments callback API (调用上面的Instruments的方法实现)
    '''

    rpc_name_prefix = "ins."
    CommandTimeout = 90 
       
    def __init__(self, rpc_server, device_id):
        self.server = rpc_server
        self.udid = device_id
        self.instruments = Instruments(self.udid)
      
    @rpc_method
    def notify(self, json_data):
        '''notify
        
        :param json_data: 要通知的内容
        :type json_data: Json
        '''
        self.instruments.notify(json_data)
    
    @rpc_method
    def send_result(self, result):
        '''发送结果给指定的设备
        
        :param device_udid: 设备的udid标识
        :type device_udid: str
        :param result: 要发送的结果
        :type result: str or dict
        '''
        self.instruments.send_result(result)
    
    @rpc_method
    def send_result_and_get_next(self, result, timeout=CommandTimeout):
        '''发送结果给指定设备，并获取下一条指令
        
        :param result: 要发送的结果
        :type result: list or dict
        :param timeout: 指令执行超时 (秒)
        :type timeout: int
        :returns : Json -- 下一条执行的指令
        '''
        return self.instruments.send_result_and_get_next(result, timeout)

    @rpc_method
    def set_command_timeout(self, timeout=CommandTimeout):
        '''设置指令执行超时
        
        :param timeout: 指令执行超时
        :type timeout: int
        '''
        self.__class__.CommandTimeout = timeout
        
    @rpc_method
    def is_working(self):
        '''判断instrument是否在工作
        
        '''
        return self.instruments.is_working()


# -*- coding:utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''utility module
'''



from testbase.util import Timeout as BaseTimeout


class Rectangle(object):
    '''UI控件的矩形区域
    '''
    def __init__(self, left, top, right, bottom):
        self.top = top
        self.left = left
        self.bottom = bottom
        self.right = right
    
    @property
    def width(self):
        return self.right - self.left
    
    @property
    def height(self):
        return self.bottom - self.top
    
    def __repr__(self):
        return repr({'top'    : self.top,
                     'left'   : self.left,
                     'bottom' : self.bottom,
                     'right'  : self.right,
                     'width'  : self.width,
                     'height' : self.height})

class EnumDirect(object):
    '''方向
    '''
    Left, Right, Up, Down = ('Left', 'Right', 'Up', 'Down')


class Timeout(BaseTimeout):
    '''超时
    '''
    def __init__(self, timeout=1.8, interval=0.005):
        BaseTimeout.__init__(self, timeout, interval)
        self.timeout = float(timeout)
        self.interval = float(interval)

    def __repr__(self):
        return "%s" % {"timeout": self.timeout, "interval": self.interval}


class RegExpCompile(object):
    '''编译正则表达式
    当字符串变量中出现正则中的逻辑运算符时，该类将会自动处理转义。
    '''
    
    def __init__(self, source):
        self.source = source
        self.__compile__()
    
    def __compile__(self):
        self.regexp = ''
        for item in self.source:
            if item in '^$.*+?=!:|\\/()[]{}':
                self.regexp += '\\'
            self.regexp += item
    
    def __repr__(self):
        return self.regexp


def less_to(version_a, version_b):
    '''判断version_a是否小于version_b
    
    :returns: True if version_a less than version_b, False if not
    :rtype: boolean
    '''
    import re
    if not re.match('\d*\.\d*\.\d*', version_a):
        return True
    elif not re.match('\d*\.\d*\.\d*', version_b):
        return False
    a = version_a.split('.')
    b = version_b.split('.')
    length = min(len(a), len(b))
    for i in range(length):
        if int(a[i]) < int(b[i]):
            return True
    return False



# -*- coding:utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''iOS App
'''

import time
import datetime

from testbase import logger


class App(object):
    '''iOS App基类
    '''
    
    def __init__(self, device, bundle_id, trace_template=None, trace_output=None, **params):
        '''构造函数
        
        :param device: Device的实例对象
        :type device: Device
        :param bundle_id: APP的BundleID（例如：com.tencent.demo）
        :type bundle_id: str
        :param trace_template: trace模板（专项测试使用，功能测试默认为None即可）
        :type trace_template: str
        :param trace_output: trace存储路径（专项测试使用，功能测试默认为None即可）
        :type trace_output: str
        :param params: app启动参数，默认为空
        :type params: dict
        '''
        self._device = device
        self._driver = device.driver
        self._bundle_id = bundle_id
        self._trace_template = trace_template
        self._trace_output = trace_output
        self._app_params = params
        self._rules_of_alert_auto_handle = []
        self._flag_alert_auto_handled = False
        self._app_started = False

    def start(self):
        '''启动APP
        '''
        begin_time = time.time()
        env = {
            'rules_of_alert_auto_handle' : self._rules_of_alert_auto_handle,
            'flag_alert_auto_handled'    : not self._flag_alert_auto_handled
        }
        self._driver.web.release_app_session(self._bundle_id) #重启app后app_id会发生变化,需要重新获取app_id
        self._app_started = self._device.start_app(self._bundle_id, self._app_params, env, self._trace_template, self._trace_output)
        if not self._app_started: raise Exception('APP-StartError')
        logger.info('[%s] APP - Start - %s - 启动耗时%s秒' % (datetime.datetime.fromtimestamp(time.time()), self._bundle_id, round(time.time() - begin_time, 3)))
    
    def get_text(self, text):
        '''获取text对应的本地语言文本
        
        :param text: 标准文本，不随语言环境发生变化的唯一标识
        :type text: str
        :return: str - 本地语言的文
        '''
        # 目前未拿到国际化语言对照表的范本，暂未实现解析步骤，用户可在子类中重载实现
        raise Exception('该函数暂未实现，请在子类中重载实现')
    
    @property
    def device(self):
        '''返回app所在的设备
        
        :rtype: qt4i.device.Device
        '''
        return self._device
    
    @property
    def language(self):
        '''app的当前语言
        
        :rtype: str
        '''
        raise NotImplementedError
    
    @property
    def driver(self):
        '''返回app所使用的driver
        
        :rtype: RPCClientProxy
        '''
        return self._driver
    
    @property
    def rules_of_alert_auto_handle(self):
        '''获取已设置的自动处理Alert规则
        
        :rtype: list - [ {"message_text": "message_text", "button_text": "button_text"}, ... ]
        '''
        return self._rules_of_alert_auto_handle

    @rules_of_alert_auto_handle.setter
    def rules_of_alert_auto_handle(self, rules=list()):
        '''设置自动处理Alert规则
        
        当message_text匹配到Alert框内的任意文本，则点击button_text指定的按钮。
        文本以段为单元，元素的label、name、value分为三段文本。规则匹配范围是Alert框内的所有元素的文本。
        :param rules: 规则
        :type rules: list - [ {"message_text": "message_text", "button_text": "button_text"}, ... ]
        '''
        self._rules_of_alert_auto_handle = rules
        if self._app_started:
            self._driver.device.set_alert_auto_handling_rules(rules)

    def add_rule_of_alert_auto_handle(self, message_text, button_text):
        '''自动处理Alert规则，新增一项
        
        :param message_text: Alert内文本片段，支持正则表达式（范围是所谓文本，每个元素的label/name/value为文本段）
        :type message_text: str
        :param button_text: Alert内按钮的文本，支持正则表达式（元素的label/name/value）
        :type button_text: str
        '''
        self._rules_of_alert_auto_handle.append({'message_text': message_text, 'button_text': button_text})
        if self._app_started:
            self._driver.device.add_alert_auto_handling_rule(message_text, button_text)

    @property
    def flag_alert_auto_handled(self):
        '''自动关闭Alert框
        :rtype: boolean
        '''
        return self._flag_alert_auto_handled
    
    @flag_alert_auto_handled.setter
    def flag_alert_auto_handled(self, flag):
        '''自动关闭Alert框
        
        :param flag: True为自动关闭|False为不自动处理
        :type  flag: boolean
        '''
        if flag and not self._flag_alert_auto_handled:
            self._flag_alert_auto_handled = True
            if self._app_started:
                self._driver.device.enable_auto_close_handling(self._device.udid)
        if not flag and self._flag_alert_auto_handled:
            self._flag_alert_auto_handled = False
            if self._app_started:
                self._driver.device.disable_auto_alert_handling(self._device.udid)
    
    def release(self):
        '''终止APP
        '''
        logger.info('[%s] APP - Release - %s' % (datetime.datetime.fromtimestamp(time.time()), self._bundle_id))
        self._app_started = False
        
        
class Safari(App):
    '''Safari浏览器
    '''
    
    
    def __init__(self, device):
        App.__init__(self, device, 'com.apple.safari', app_name='Safari')
        self._init_window()
        
    def _init_window(self):
        from qt4i.icontrols import Element, Window
        from qt4i.qpath import QPath
        self._win = Window(self)
        self._win.updateLocator({
            'Safari':{'type':Element, 'root':self, 'locator':"Safari"},       
            'url输入框.未激活键盘': {'type':Element, 'root':self, 'locator':QPath("/name='URL' & maxdepth=9 & instance=1")},
            'url输入框': {'type':Element, 'root':self, 'locator':QPath("/name~='地址|Address|URL' & maxdepth=9 & instance=1")},
            '打开url': {'type':Element, 'root':self, 'locator':QPath("/name~='^打开$|^Open$' & maxdepth=15")},
        })
        self._win.Controls['Safari'].click()
    
    def open_url(self, url, timeout=1):
        '''打开Safari浏览器，跳转指定网页
        
        :param url: url地址
        :type url: str
        '''
        if self._win.Controls['url输入框.未激活键盘'].wait_for_exist(timeout, 0.05):
            self._win.Controls['url输入框.未激活键盘'].click()
        self._win.Controls['url输入框'].value = url
        self._win.Controls['url输入框'].send_keys('\n')
        if self._win.Controls['打开url'].exist():
            open_btn = self._win.Controls['打开url']
            open_btn.click()
            if open_btn.exist(): # 补充点击防止点击不生效
                self.device.click2(open_btn)
        else:
            raise Exception("Safari没有出现\"打开\"的对话框")


class NLCType(object):
    '''模拟弱网络类型
    '''
    
    NONE = "None"
    LOSS = "100%Loss"
    DSL = "DSL"
    _3G = "3G"
    EDGE = "Edge"
    DNS = "High Latency DNS"
    LTE = "LTE"
    BADNET = "Very Bad Network"
    WIFI = "Wi-Fi"
    WIFI802 = "WI-Fi 802.11ac"
    ADD = "ADD"
    

class Preferences(App):
    '''系统app 设置
    '''
    
    
    def __init__(self, device):
        App.__init__(self, device, 'com.apple.Preferences', app_name='设置')
        self._init_window()
        
    def _init_window(self):
        from qt4i.icontrols import Element, Window
        from qt4i.qpath import QPath
        self._win = Window(self)
        self._win.updateLocator({
            '无线局域网':{'type':Element, 'root':self, 'locator':QPath("/classname='Cell' && name='无线局域网' && maxdepth=8")},
            '无线开关':{'type':Element, 'root':self, 'locator':QPath("/classname='Switch' && name='无线局域网' && maxdepth=12")},
            '设置':{'type':Element, 'root':self, 'locator':QPath("/classname='Button' && name='设置' && maxdepth=5")},
            '蜂窝移动网络':{'type':Element, 'root':self, 'locator':QPath("/classname='Cell' && name='蜂窝移动网络' && maxdepth=8")},
            '蜂窝开关':{'type':Element, 'root':self, 'locator':QPath("/classname='Switch' && name='蜂窝移动数据' && maxdepth=12")},
            '飞行开关':{'type':Element, 'root':self, 'locator':QPath("/classname='Switch' && name='飞行模式' && maxdepth=9")},
            
            '开发者':{'type':Element, 'root':self, 'locator':QPath("/classname='Cell' && name='开发者' && maxdepth=8")},
            'status':{'type':Element, 'root':self, 'locator':QPath("/classname='Cell' && name='Status' && maxdepth=11")},
            'Enable':{'type':Element, 'root':self, 'locator':QPath("/classname='Switch' && name='Enable' && maxdepth=12")},
            '100%Loss':{'type':Element, 'root':self, 'locator':QPath("/classname='Cell' && name='100% Loss' && maxdepth=11")},
            '3G':{'type':Element, 'root':self, 'locator':QPath("/classname='Cell' && name='3G' && maxdepth=11")},
            'DSL':{'type':Element, 'root':self, 'locator':QPath("/classname='Cell' && name='DSL' && maxdepth=11")},
            'Edge':{'type':Element, 'root':self, 'locator':QPath("/classname='Cell' && name='Edge' && maxdepth=11")},
            'High Latency DNS':{'type':Element, 'root':self, 'locator':QPath("/classname='Cell' && name='High Latency DNS' && maxdepth=11")},
            'LTE':{'type':Element, 'root':self, 'locator':QPath("/classname='Cell' && name='LTE' && maxdepth=11")},
            'Very Bad Network':{'type':Element, 'root':self, 'locator':QPath("/classname='Cell' && name='Very Bad Network' && maxdepth=11")},
            'Wi-Fi':{'type':Element, 'root':self, 'locator':QPath("/classname='Cell' && name='Wi-Fi' && maxdepth=11")},
            'WI-Fi 802.11ac':{'type':Element, 'root':self, 'locator':QPath("/classname='Cell' && name='Wi-Fi 802.11ac' && maxdepth=11")},
            'ADD':{'type':Element, 'root':self, 'locator':QPath("/classname='Cell' && name='ADD' && maxdepth=11")},
            '返回':{'type':Element, 'root':self, 'locator':QPath("/classname='Button' && name='返回' && maxdepth=5 && instance=1")},      
            #ios10  
            '手动':{'type':Element, 'root':self, 'locator':"手动"},   
            '关闭':{'type':Element, 'root':self, 'locator':"关闭"},  
            '服务器':{'type':Element, 'root':self, 'locator':QPath("/classname = 'TextField' & label = '服务器' & maxdepth = 11")},
            '端口':{'type':Element, 'root':self, 'locator':QPath("/classname = 'TextField' & label = '端口'  & maxdepth = 11")},              
            #ios11
            '配置代理':{'type':Element, 'root':self, 'locator':"配置代理"},         
            '存储':{'type':Element, 'root':self, 'locator':"存储"},   
            '返回.无线局域网':{'type':Element, 'root':self, 'locator':"无线局域网"},     
            'Conditioner':{'type':Element, 'root':self, 'locator':"Network Link Conditioner"}, 
        })
        self.start()
             
    def _switch_wifi(self, network_type, timeout=1):
        '''wifi切换
        '''
        #进入无线网络
        self._win.Controls['无线局域网'].click()
        #拨动开关按钮        
        if int(self._win.Controls['无线开关'].value) == 1:
            if network_type == 0 or network_type == 1:
                self._win.Controls['无线开关'].click()
        else:
            if network_type == 2 or network_type == 3:
                self._win.Controls['无线开关'].click()
        #返回设置页面
        self._win.Controls['设置'].click()
        
    def _switch_xG(self, network_type, timeout=1):
        '''xG网络切换
        '''
        #进入蜂窝移动网络
        self._win.Controls['蜂窝移动网络'].click()
        #拨动开关按钮               
        if int(self._win.Controls['蜂窝开关'].value) == 1:
            if network_type == 0 or network_type == 2:
                self._win.Controls['蜂窝开关'].click()
        else:
            if network_type == 1 or network_type == 3:
                self._win.Controls['蜂窝开关'].click()
        #返回设置页面
        self._win.Controls['设置'].click()     
        
    def _switch_nlc(self, nlc_type=NLCType.NONE, timeout=1):
        '''弱网络模拟器切换
        '''
        #进入开发者模式
        self._win.Controls['开发者'].click()
        #获得版本号
        version = int(self._device.ios_version[0])
        if version == 1 :
            version = int(self._device.ios_version[:2])  
        #进入状态选择 
        if version == 11 :
            self._win.Controls['Conditioner'].click()
        else :
            self._win.Controls['status'].click()
        status = self._win.Controls['Enable']
        if nlc_type == NLCType.NONE:
            if int(status.value) == 1:
                status.click()
        else :
            if int(status.value) == 0:
                status.click()
            self._win.Controls[nlc_type].click()
        if version == 11 :
            pass
        else :
            self._win.Controls['返回'].click()
            self._win.Controls['设置'].click()           
         
    def switch_network(self, network_type, nlc_type=NLCType.NONE, timeout=1):
        '''网络切换
        
        :param network_type: 网络类型，如下：
                                     0:无WIFI无4G
                                     1:无WIFi有4G
                                     2:有WIFI无4G
                                     3:有WIFI有4G
                                     4:飞行模式
                                     5:保持不变,仅设置弱网
        :type network_type: int
        :param nlc_type: 模拟弱网络类型
        :type nlc_type: NLCType
        '''
        if network_type == 4:
            value = int(self._win.Controls['飞行开关'].value)
            if value == 0 :
                self._win.Controls['飞行开关'].click()
        elif network_type == 5:
            value = int(self._win.Controls['飞行开关'].value)
            if value == 1 :
                self._win.Controls['飞行开关'].click()
            self._switch_nlc(nlc_type, timeout)
        else:
            if self._win.Controls['飞行开关'].value == True:
                self._win.Controls['飞行开关'].click()
            self._switch_wifi(network_type, timeout)
            self._switch_xG(network_type, timeout)
            self._switch_nlc(nlc_type, timeout)
        
    def set_host_proxy(self, server, port, wifi_name): 
        '''设置host代理
        
        :param server: 服务器名
        :type server: str
        :param port: 端口号
        :type port: int
        :param wifi: wifi名
        :type wifi: str
        '''
        #ios版本
        version = int(self._device.ios_version[0])
        if version == 1 :
            version = int(self._device.ios_version[:2])   
        time.sleep(1)                            #添加保护
        #进入无线局域网
        if self._win.Controls['无线局域网'].wait_for_exist(2,0.05):
            self._win.Controls['无线局域网'].click()
        time.sleep(1)                            #添加保护
        from qt4i.icontrols import Element
        from qt4i.qpath import QPath
        self._win.updateLocator({
            'wifi_name':{'type':Element, 'root':self, 'locator':wifi_name},       
            'wifi_title':{'type':Element, 'root':self, 
                'locator':QPath("/classname='NavigationBar' & maxdepth=4 & name='%s'" % wifi_name)},       
            })
        self._win.Controls['wifi_name'].click()
        time.sleep(1)                            #添加保护
        if self._win.Controls['wifi_title'].exist():
            logger.info('已经是%s, 无需切换WiFi' % wifi_name)
        else:
            #再点击一次进入代理设置
            self._win.Controls['wifi_name'].click()
        
        if version == 11 :
            self._win.Controls['配置代理'].click()
            
        self._win.Controls['手动'].click()
        #服务器框
        server_text_field = self._win.Controls['服务器']
        server_text_field.click()
        server_text_field.value = server + '\n'
        #端口框
        port_text_field = self._win.Controls['端口']
        port_text_field.value = port
    
        if version == 11 :
            self._win.Controls['存储'].click()    
            if self._win.Controls['wifi_name'].wait_for_exist(2,0.05):
                self._win.Controls['wifi_name'].click()
        self._win.Controls['返回.无线局域网'].click()
        self._win.Controls['设置'].click()
        
    def reset_host_proxy(self):   
        '''关闭host代理
        ''' 
        #ios版本
        version = int(self._device.ios_version[0])
        if version == 1 :
            version = int(self._device.ios_version[:2])  
        time.sleep(1)                            #添加保护
        #进入无线局域网
        if self._win.Controls['无线局域网'].wait_for_exist(2,0.05):
            self._win.Controls['无线局域网'].click()
        time.sleep(1)                            #添加保护
        from qt4i.icontrols import Element
        self._win.updateLocator({
            'wifi_name':{'type':Element, 'root':self, 'locator':self._device.wifi},       
            })
        self._win.Controls['wifi_name'].click()
        time.sleep(1)                            #添加保护
        #再点击一次进入代理设置
        self._win.Controls['wifi_name'].click()
        
        if version == 11 :
            self._win.Controls['配置代理'].click()
        
        self._win.Controls['关闭'].click()
        if version == 11 :
            self._win.Controls['存储'].click()    
            if self._win.Controls['wifi_name'].wait_for_exist(2,0.05):
                self._win.Controls['wifi_name'].click()
        self._win.Controls['返回.无线局域网'].click()
        self._win.Controls['设置'].click()
        


# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''QT4i Driver Server
'''

from __future__ import absolute_import, print_function

import argparse
import atexit
import os
import signal
import sys
import time
from six import string_types
    
from qt4i.driver.rpc import SimpleJSONRPCServer
from qt4i.driver.tools import logger as logging 
from qt4i.driver.util import Process

DEFAULT_IP = '0.0.0.0'
DEFAULT_PORT = 12306
DEFAULT_AGENT_PORT = 8100
DEFAULT_PID_FILE = '/tmp/driverserver.pid'

logger = logging.get_logger()

class ArgumentParser(argparse.ArgumentParser):
    '''Argument parser, automatically display help while error
    '''
    def error(self, message):
        sys.stderr.write('error: %s\n' % message)
        self.print_help()
        sys.exit(2) 
CMD_HELP = {
    "start": "start driver server daemon",
    "stop" :  "stop driver server daemon",
    "restart": "stop driver server daemon and start it again",
    "debug": "try to start driver server as a normal process for debug",
}

#参数解析器
main_parser = ArgumentParser(description='QT4i 3.0 driver program',
                                 formatter_class=argparse.RawDescriptionHelpFormatter )
main_parser.add_argument('--host', '-H', metavar='HOST', dest='host', default=DEFAULT_IP,
                         help='listening address of driver server, local address<%s> is used by default' % DEFAULT_IP)
main_parser.add_argument('--port', '-p', metavar='PORT', dest='port', default=DEFAULT_PORT,
                         help='listening port of driver server, port %s is used by default' % DEFAULT_PORT)
main_parser.add_argument('--type', '-t', metavar='TYPE', dest='driver_type', default="instruments",
                         help='driver type of driver server(instruments or xctest), "instruments" is used by default')
main_parser.add_argument('--pidfile', '-f', metavar='PIDFILE', dest='pidfile', default=DEFAULT_PID_FILE,
                         help='the path of pid file, %s is used by default' % DEFAULT_PID_FILE)
main_parser.add_argument('--udid', '-u', metavar='UDID', dest='udid', default=None,
                         help='udid of device')
main_parser.add_argument('--agent_port', '-a', metavar='AGENTPORT', dest='agent_port', default=DEFAULT_AGENT_PORT,
                         help='listening port of xctest agent, port %s is used by default' % DEFAULT_AGENT_PORT)
main_parser.add_argument('--web_port', '-w', metavar='WEBPORT', dest='web_port', default=None,
                         help='listening port of simulator for webview testing')
main_parser.add_argument('--register', '-r', dest='endpoint_clss', default=None,
                         help='the endpoint classes registered to driverserver, for example, "perflib.ios.fps.FPSMonitor,perflib.ios.leaks.LeaksMonitor"')
     
subparser_dict = {}
subparsers = main_parser.add_subparsers(title='List of Commands', metavar='COMMAND')
for command in CMD_HELP:
    subparser_dict[command] = subparsers.add_parser(command, help=CMD_HELP[command])
    subparser_dict[command].set_defaults(func=command)

subparser_dict['stop'].add_argument('--force', '-f', action='store_true', default=False, 
                                    help='force stop driver server')

class Daemon(object):
    """
    A generic daemon class.
    Usage: subclass the Daemon class and override the run() method
    """

    def __init__(self, pidfile, stdin='/dev/null', stdout='/dev/null', stderr='/dev/null'):
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.pidfile = pidfile

    def daemonize(self):
        """
        do the UNIX double-fork magic, see Stevens' "Advanced 
        Programming in the UNIX Environment" for details (ISBN 0201563177)
        http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
        """
        # Do first fork
        self.fork()

        # Decouple from parent environment
        self.dettach_env()

        # Do second fork
        self.fork()

        # Flush standart file descriptors
        sys.stdout.flush()
        sys.stderr.flush()

        # 
        self.attach_stream('stdin', mode='r')
        self.attach_stream('stdout', mode='a+')
        self.attach_stream('stderr', mode='a+')
       
        # write pidfile
        self.create_pidfile()

    def attach_stream(self, name, mode):
        """
        Replaces the stream with new one
        """
        stream = open(getattr(self, name), mode)
        os.dup2(stream.fileno(), getattr(sys, name).fileno())

    def dettach_env(self):
        os.chdir("/")
        os.setsid()
        os.umask(0)

    def fork(self):
        """
        Spawn the child process
        """
        try:
            pid = os.fork()
            if pid > 0:
                sys.exit(0)
        except OSError as e:
            sys.stderr.write("Fork failed: %d (%s)\n" % (e.errno, e.strerror))
            sys.exit(1)

    def create_pidfile(self):
        atexit.register(self.delpid)
        pid = str(os.getpid())
        open(self.pidfile,'w+').write("%s\n" % pid)

    def delpid(self):
        """
        Removes the pidfile on process exit
        """
        os.remove(self.pidfile)

    def start(self, is_daemon = True):
        """
        Start the daemon
        """
        # Check for a pidfile to see if the daemon already runs
        pid = self.get_pid()

        if pid:
            if Process().get_process_by_pid(pid, 'driverserver.py'):
                message = "pidfile %s already exist. Daemon already running?\n"
                logger.warning(message % self.pidfile)
                return
            else:
                os.remove(self.pidfile)

        # Start the daemon
        if is_daemon:
            self.daemonize()
        self.run()

    def get_pid(self):
        """
        Returns the PID from pidfile
        """
        try:
            pf = open(self.pidfile,'r')
            pid = int(pf.read().strip())
            pf.close()
        except (IOError, TypeError):
            pid = None
        return pid

    def stop(self, silent=False):
        """
        Stop the daemon
        """
        # Get the pid from the pidfile
        pid = self.get_pid()

        if not pid:
            if not silent:
                message = "pidfile %s does not exist. Daemon not running?\n"
                logger.error(message % self.pidfile)
            return 

        # Try killing the daemon process    
        try:
            while True:
                os.kill(pid, signal.SIGTERM)
                time.sleep(0.1)
        except OSError as err:
            err = str(err)
            if err.find("No such process") > 0:
                if os.path.exists(self.pidfile):
                    os.remove(self.pidfile)
            else:
                logger.error(str(err))
                sys.exit(1)

    def restart(self):
        """
        Restart the daemon
        """
        self.stop(silent=True)
        self.start()

    def run(self):
        """
        You should override this method when you subclass Daemon. It will be called after the process has been
        daemonized by start() or restart().
        """
        raise NotImplementedError

class DriverManager(Daemon):
    
    def __init__(self, pidfile, urls, address, port, udid, agent_port, web_port):
        '''构造（原则：Driver的配置只在此处配置）
        '''
        Daemon.__init__(self, pidfile)
        self._urls = urls
        self._driver_address = address
        self._driver_port = port
        self._driver_url = 'http://%s:%d' % (self._driver_address, self._driver_port)
        self._udid = udid
        self._agent_port = agent_port
        if isinstance(web_port, string_types):
            web_port = int(web_port)
        self._web_port = web_port
    
    @property
    def driver_url(self):
        return self._driver_url
    
    def start(self, is_daemon = True):
        '''启动driver
        '''
        # Check for a pidfile to see if the daemon already runs
        pid = self.get_pid()

        if pid:
            process = Process().get_process_by_pid(pid, 'driverserver.py')
            if process:
                args = process['CMD'].split(' ')
                if len(args) > 3:
                    args = main_parser.parse_args(args[2:])
                    logger.info('Driverserver info:%s' % args)
                    if int(args.agent_port) == self._agent_port and int(args.port) == self._driver_port:
                        return
                Process().kill_process_by_pid(pid)
                logger.warning('Kill old driver server process %d'  % pid)
                logger.warning('old_params: driver_port: %d, agent_port: %d' % (int(args.port), int(args.agent_port)))
                logger.warning('new_params: driver_port: %d, agent_port: %d' % (self._driver_port, self._agent_port))
            else:
                os.remove(self.pidfile)

        # Start the daemon
        if is_daemon:
            self.daemonize()
        self.run()
        
    def run(self):
        if self._udid:
            if self._web_port:
                # 单例类PortManager，设置WebInspector端口
                from qt4i.driver.tools.sched import PortManager
                PortManager.set_port(port_type='web', udid=self._udid, port=self._web_port, base=27753)
            # 单例类XCUITestAgentManager，启动XCUITestAgent
            from qt4i.driver.xctest.agent import XCUITestAgentManager
            agent_manager = XCUITestAgentManager()
            agent_manager.start_agent(device_id=self._udid, server_port=self._agent_port, retry=1)
        
        server = SimpleJSONRPCServer(urls=urls, addr=(self._driver_address, self._driver_port))
        logger.info('DriverServer(%s:%s) - started' % (self._driver_address, self._driver_port))
        server.serve_forever()
        
    def cleanup(self):
        '''清理环境
        device中的stop_all_agents接口需要udid，请使用host中的对应接口
        '''
        from qt4i.driver.rpc import RPCClientProxy
        if self._udid:
            driver = RPCClientProxy('%s/device/%s/' % (self._driver_url, str(self._udid)), allow_none=True, encoding='UTF-8')
            driver.device.stop_all_agents()
        else:
            host = RPCClientProxy('%s/host/' % (self._driver_url), allow_none=True, encoding='UTF-8')
            host.stop_all_agents()
    
    def stop(self, silent=False):
        '''停止driver
        '''
        try:
            self.cleanup()
        except:
            logger.exception('Stop Driver Server')
        Daemon.stop(self, silent)
    
def load(cls_name):
    '''加载对应的对象
    
    :param cls_name: 注册rpc类名，例如：perflib.ios.fps.FPSMonitor
    :type cls_name: string
    :returns - Type/ModuleType
    '''
    try:
        parts = cls_name.split('.')
        module = None
        parts_imp = parts[:]
        while parts_imp:
            try:
                modulename = '.'.join(parts_imp)
                __import__(modulename)   #__import__得到的是最外层模块的object
                module = sys.modules[modulename]
                break
            except ImportError:
                del parts_imp[-1]
            except:
                del parts_imp[-1]
                break
        if parts_imp == parts: #为一个包或模块
            raise RuntimeError(" %s 是一个包或模块，请注册一个类" %cls_name)
        elif parts_imp == parts[0:-1] and hasattr(module, parts[-1]): #为一个类
            endpoint_cls = getattr(module, parts[-1]) 
            return endpoint_cls
        else: #触发异常
            if parts_imp:
                modulename = '.'.join(parts_imp)
                modulename += '.'
            else:
                modulename = ''
            modulename += parts[len(parts_imp)]
            __import__(modulename)
    except:
        logger.exception('load %s' %cls_name)
        raise

if __name__ == '__main__':
    args = main_parser.parse_args()
    urls = []
    if args.driver_type == 'instruments':
        from qt4i.driver.instruments.uia import Device, Application, Element #@UnusedImport
        from qt4i.driver.instruments.ins import InstrumentsCallBacker 
        urls = [
            ("^device/(?P<device_id>[\w\-]+)/$", "^ins\..*", InstrumentsCallBacker)
        ]
    else:
        from qt4i.driver.xctest.wda import Device, Application, Element 
    
    from qt4i.driver.host import RPCServerHost
    from qt4i.driver.web.web import WebInspector
    urls.extend([("^host/$", None, RPCServerHost),
            ("^device/(?P<device_id>[\w\-]+)/$", "^web\..*", WebInspector),
            ("^device/(?P<device_id>[\w\-]+)/$", "^device\..*", Device),
            ("^device/(?P<device_id>[\w\-]+)/$", "^app\..*", Application),
            ("^device/(?P<device_id>[\w\-]+)/$", "^element\..*", Element),])
    
    if args.endpoint_clss:
        endpoint_clss = args.endpoint_clss.split(",")
        for cls in endpoint_clss:
            if cls:
                endpoint_cls = load(cls)
                cls_name = endpoint_cls.__name__
                urls.append(("^device/(?P<device_id>[\w\-]+)/$", "^%s\..*" % cls_name.lower(), endpoint_cls))

    dm = DriverManager(args.pidfile, urls, args.host, int(args.port), args.udid, int(args.agent_port), args.web_port)
    if args.func == 'start':
        dm.start()
          
    elif args.func == 'stop':
        dm.stop()
          
    elif args.func == 'restart':
        dm.restart()
              
    elif args.func == 'debug':
        dm.start(False)
          
    else:
        main_parser.print_help()
        sys.exit(1)



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''qt4i测试用例基类模块
'''

from __future__ import absolute_import, print_function

import os
import time
import traceback

from testbase.testcase import TestCase
from qt4i.device import Device, QT4i_LOGS_PATH


class iTestCase(TestCase):
    '''QT4i测试用例基类
    '''
    crash_flag = False
    attachments_path = QT4i_LOGS_PATH
    error_print_uitree = False
    
    def init_test(self, testresult):
        '''测试用例初始化
        '''
        super(iTestCase, self).init_test(testresult)
        self._start_time = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())

    initTest = init_test    #兼容驼峰式命名和下划线式命名
    
    def clean_test(self):
        '''测试用例清理，以用例为单位清理日志，并释放所有设备
        '''
        try:
            super(iTestCase, self).clean_test()
            Device.cleanup_all_log()
        finally:
            Device.release_all()
    
    cleanTest = clean_test  #兼容驼峰式命名和下划线式命名

    def get_extra_fail_record(self):
        '''当错误发生时，获取需要额外添加的日志记录和附件信息
        
        :return: dict,dict - 日志记录，附件信息
        '''
        #（instruments截屏失败后 则会使用libimobiledevice截屏）
        attachments = {}
        try:
            for device in Device.Devices:
                #上传截图到报告
                image_path = os.path.join(self.attachments_path, "%s_%s.png" % (type(self).__name__, time.time()))
                if device.screenshot(image_path)[0]:
                    attachments['设备截图(%s)' % device.name] = image_path
    
                #上传日志到报告
                log_path = os.path.join(self.attachments_path, "%s_%s.log" % (type(self).__name__, time.time()))
                with open(log_path, 'w') as fd:
                    fd.write(device.get_log(self._start_time))
                    attachments['日志(%s)' % device.name] = log_path
                
                #打印发生error时的UI tree到报告中
                if self.error_print_uitree: 
                    self.log_info("error time ui tree: \n %s"%'\n'.join(device.print_uitree(self.error_print_uitree)))
    
                #记录当前是否发生crash,在具体的测试基类中（根据进程名称过滤）上传crash日志
                self.crash_flag = device.driver.app.has_crashed()
    
                #上传driver日志到报告
                driver_log_path = os.path.join(self.attachments_path, "driver%s_%s.log" % (type(self).__name__, time.time()))
                with open(driver_log_path, 'w') as fd:
                    fd.write(device.get_driver_log(self._start_time))
                    attachments['driver日志(%s)' % device.name] = driver_log_path
            self._start_time = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())
        except:
            self.log_info(traceback.format_exc())
        finally:
            return {}, attachments
    
    def get_crash_log(self, procname):
        '''获取应用程序的crash日志
        
        :param procname: app的进程名，可通过xcode查看
        :type procname: str
        :return: string or None - crash日志路径
        '''
        crash_logs = {}
        for device in Device.Devices:
            crash_log = device.get_crash_log(procname)
            if crash_log:
                crash_logs['crash日志(%s)' % device.name] = crash_log
        return crash_logs 



# -*- coding:utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''UIAutomation和XCTest的控件映射表
'''
import six

from tuia.qpathparser import QPathParser


XCT_UIA_MAPS = {
    'Any': 'UIAElement',
    'Other': 'UIAElement',
    'Application':'UIAApplication',
    'Group': 'UIATableGroup',
    'Window': 'UIAWindow',
    'Sheet': 'UIAActionSheet',
    'Alert': 'UIAAlert',
    'Button': 'UIAButton',
    'Keyboard': 'UIAKeyboard',
    'Key': 'UIAKey',
    'NavigationBar': 'UIANavigationBar',
    'TabBar': 'UIATabBar',
    'Toolbar': 'UIAToolbar',
    'StatusBar': 'UIAStatusBar',
    'Table': 'UIATableView',
    'CollectionView': 'UIACollectionView',
    'Slider': 'UIASlider',
    'PageIndicator': 'UIAPageIndicator',
    'ProgressIndicator': 'UIAProgressIndicator',
    'Popover': 'UIAPopover',
    'ActivityIndicator': 'UIAActivityIndicator',
    'SegmentedControl': 'UIASegmentedControl',
    'Picker': 'UIAPicker',
    'PickerWheel': 'UIAPickerWheel',
    'Switch': 'UIASwitch',
    'Link': 'UIALink',
    'Image': 'UIAImage',
    'SearchField': 'UIASearchBar',
    'ScrollView': 'UIAScrollView',
    'StaticText': 'UIAStaticText',
    'TextField': 'UIATextField',
    'SecureTextField': 'UIASecureTextField',
    'TextView': 'UIATextView',
    'Menu': 'UIAEditingMenu',
    'MenuItem': 'UIAElement',
    'Map': 'UIAMapView',
    'WebView': 'UIAWebView',
    'Cell': 'UIATableCell',
}

UIA_XCT_MAPS = {
    'UIAElement':'Other',
    'UIAApplication':'Application',
    'UIATableGroup':'Group',
    'UIAWindow':'Window',
    'UIAActionSheet':'Sheet',
    'UIAAlert':'Alert',
    'UIAButton':'Button',
    'UIAKeyboard':'Keyboard',
    'UIAKey':'Key',
    'UIANavigationBar':'NavigationBar',
    'UIATabBar':'TabBar',
    'UIAToolbar': 'Toolbar',
    'UIAStatusBar':'StatusBar',
    'UIATableView':'Table',
    'UIACollectionView':'CollectionView',
    'UIASlider':'Slider',
    'UIAPageIndicator':'PageIndicator',
    'UIAProgressIndicator':'ProgressIndicator',
    'UIAPopover':'Popover',
    'UIAActivityIndicator':'ActivityIndicator',
    'UIASegmentedControl':'SegmentedControl',
    'UIAPicker':'Picker',
    'UIAPickerWheel':'PickerWheel',
    'UIASwitch':'Switch',
    'UIALink':'Link',
    'UIAImage':'Image',
    'UIASearchBar':'SearchField',
    'UIAScrollView':'ScrollView',
    'UIAStaticText':'StaticText',
    'UIATextField':'TextField',
    'UIASecureTextField':'SecureTextField',
    'UIATextView':'TextView',
    'UIAEditingMenu':'Menu',
    'UIAMapView':'Map',
    'UIAWebView':'WebView',
    'UIATableCell':'Cell',
}

def xctest2instruments(locator):
    qpath_array, _ = QPathParser().parse(locator)
    for node in qpath_array:
        if 'classname' in node:
            if node['classname'][1] in XCT_UIA_MAPS.keys():
                node['classname'][1] = XCT_UIA_MAPS[node['classname'][1]]
    new_locator = ''   
    for node in qpath_array:
        new_locator += '/'
        properties = []
        for p in node.keys():
            if isinstance(node[p][1], six.string_types):
                value = node[p][1]
                if six.PY2 and isinstance(node[p][1], unicode):
                    value = value.encode('utf-8')
                node_str = "{}{}\'{}\'".format(p, node[p][0], value)   
            else:
                node_str = "{}{}{}".format(p, node[p][0], node[p][1]) 
            properties.append(node_str)
        properties = ' & '.join(properties)
        new_locator += properties.decode('utf-8')
    return new_locator
